
from __future__ import annotations

import asyncio
import logging
import time

from typing import Dict, Iterable, List, Optional, Tuple

from qtoggleserver import persist
from qtoggleserver import system
from qtoggleserver.conf import settings
from qtoggleserver.core import events as core_events
from qtoggleserver.core import ports as core_ports
from qtoggleserver.core.typing import GenericJSONDict, PortValue
from qtoggleserver.utils import json as json_utils


PERSIST_COLLECTION = 'value_history'

_CACHE_TIMESTAMP_MIN_AGE = 3600 * 1000  # Don't cache samples newer than this number of milliseconds ago

logger = logging.getLogger(__name__)

_history_event_handler: Optional[HistoryEventHandler] = None
_sampling_task: Optional[asyncio.Task] = None
_janitor_task: Optional[asyncio.Task] = None

# Samples cached by port_id and timestamp
_samples_cache: Dict[str, Dict[int, PortValue]] = {}

# Used to schedule sample removal with remove_samples(..., background=True)
_pending_remove_samples: List[Tuple[core_ports.BasePort, Optional[int], Optional[int]]] = []


class HistoryEventHandler(core_events.Handler):
    FIRE_AND_FORGET = True

    async def handle_event(self, event: core_events.Event) -> None:
        if not isinstance(event, core_events.ValueChange):
            return  # We're only interested in port value changes

        if not system.date.has_real_date_time():
            return  # Don't record history unless we've got real date/time

        port = event.get_port()
        history_interval = await port.get_history_interval()
        if history_interval != -1:
            return  # Only consider ports with history interval set to special -1 (on value change)

        now_ms = int(time.time() * 1000)

        await save_sample(port, now_ms)
        port.set_history_last_timestamp(now_ms)


async def sampling_task() -> None:
    while True:
        try:
            await asyncio.sleep(1)

            if not system.date.has_real_date_time():
                continue  # Don't record history unless we've got real date/time

            now_ms = int(time.time() * 1000)
            for port in core_ports.get_all():
                if not port.is_enabled():
                    continue

                history_last_timestamp = port.get_history_last_timestamp()
                history_interval = await port.get_history_interval()
                if history_interval <= 0:  # Disabled or on value change
                    continue

                if now_ms - history_last_timestamp < history_interval * 1000:
                    continue

                await save_sample(port, now_ms)
                port.set_history_last_timestamp(now_ms)
                await port.save()  # history_last_timestamp must be persisted

        except asyncio.CancelledError:
            logger.debug('sampling task cancelled')
            break

        except Exception as e:
            logger.error('sampling task error: %s', e, exc_info=True)


async def janitor_task() -> None:
    global _pending_remove_samples

    while True:
        try:
            await asyncio.sleep(settings.core.history_janitor_interval)

            if not system.date.has_real_date_time():
                continue

            now = int(time.time())
            for port in core_ports.get_all():
                history_retention = await port.get_history_retention()
                if history_retention <= 0:
                    continue

                to_timestamp = (now - history_retention) * 1000
                logger.debug('removing old samples of %s from history', port)
                await remove_samples([port], from_timestamp=0, to_timestamp=to_timestamp)

            # Remove samples that were background-scheduled for removal
            # Group together requests w/o timestamp bounds

            ports = []
            rem_pending_remove_samples = []
            for port, from_timestamp, to_timestamp in _pending_remove_samples:
                if from_timestamp is to_timestamp is None:
                    ports.append(port)

                else:
                    rem_pending_remove_samples.append((port, from_timestamp, to_timestamp))

            if ports:
                port_ids = [p.get_id() for p in ports]
                logger.debug('removing samples of %s from history (background)', ', '.join(port_ids))
                await remove_samples(ports)

            _pending_remove_samples = rem_pending_remove_samples
            while _pending_remove_samples:
                port, from_timestamp, to_timestamp = _pending_remove_samples.pop(0)
                logger.debug('removing samples of %s from history (background)', port)
                await remove_samples([port], from_timestamp=from_timestamp, to_timestamp=to_timestamp)

        except asyncio.CancelledError:
            logger.debug('janitor task cancelled')
            break

        except Exception as e:
            logger.error('janitor task error: %s', e, exc_info=True)


def is_enabled() -> bool:
    return persist.is_history_supported() and settings.core.history_support


async def get_samples_slice(
    port: core_ports.BasePort,
    from_timestamp: Optional[int] = None,
    to_timestamp: Optional[int] = None,
    limit: Optional[int] = None,
    sort_desc: bool = False
) -> Iterable[GenericJSONDict]:
    filt = {
        'pid': port.get_id(),
    }

    if from_timestamp is not None:
        filt.setdefault('ts', {})['ge'] = from_timestamp

    if to_timestamp is not None:
        filt.setdefault('ts', {})['lt'] = to_timestamp

    sort = 'ts'
    if sort_desc:
        sort = f'-{sort}'

    results = await persist.query(PERSIST_COLLECTION, filt=filt, sort=sort, limit=limit)

    return ({'value': r['val'], 'timestamp': r['ts']} for r in results)


async def get_samples_by_timestamp(
    port: core_ports.BasePort,
    timestamps: List[int]
) -> Iterable[GenericJSONDict]:
    port_filter = {
        'pid': port.get_id(),
    }

    now_ms = int(time.time() * 1000)
    samples_cache = _samples_cache.setdefault(port.get_id(), {})
    INEXISTENT = {}

    query_tasks = []
    for timestamp in timestamps:
        # Look it up in cache
        sample = samples_cache.get(timestamp, INEXISTENT)
        if sample is INEXISTENT:
            filt = dict(port_filter, ts={'le': timestamp})
            task = persist.query(PERSIST_COLLECTION, filt=filt, sort='-ts', limit=1)

        else:
            task = asyncio.Future()
            task.set_result([sample])

        query_tasks.append(task)

    task_results = await asyncio.gather(*query_tasks)

    samples = []
    for i, task_result in enumerate(task_results):
        timestamp = timestamps[i]

        query_results = list(task_result)
        if query_results:
            sample = query_results[0]
            samples.append(sample)

        else:
            samples.append(None)

        # Add sample to cache if it's old enough
        if now_ms - timestamp > _CACHE_TIMESTAMP_MIN_AGE:
            samples_cache[timestamp] = samples[-1]

    return ({'value': r['val'], 'timestamp': r['ts']} if r is not None else None for r in samples)


async def save_sample(port: core_ports.BasePort, timestamp: int) -> None:
    value = port.get_last_read_value()

    if value is None:
        logger.debug('skipping null sample of %s (timestamp = %s)', port, timestamp)
        return

    logger.debug('saving sample of %s (value = %s, timestamp = %s)', port, json_utils.dumps(value), timestamp)

    record = {
        'pid': port.get_id(),
        'val': value,
        'ts': timestamp
    }

    await persist.insert(PERSIST_COLLECTION, record)


async def remove_samples(
    ports: List[core_ports.BasePort],
    from_timestamp: Optional[int] = None,
    to_timestamp: Optional[int] = None,
    background: bool = False
) -> Optional[int]:
    filt = {
        'pid': {'in': [p.get_id() for p in ports]}
    }

    if from_timestamp is not None:
        filt.setdefault('ts', {})['ge'] = from_timestamp

    if to_timestamp is not None:
        filt.setdefault('ts', {})['lt'] = to_timestamp

    # Invalidate samples cache for the ports
    for port in ports:
        _samples_cache.pop(port.get_id(), None)

    if background:
        for port in ports:
            _pending_remove_samples.append((port, from_timestamp, to_timestamp))

    else:
        return await persist.remove(PERSIST_COLLECTION, filt)


async def reset() -> None:
    logger.debug('clearing persisted data')
    await persist.remove(PERSIST_COLLECTION)


async def init() -> None:
    global _history_event_handler
    global _sampling_task
    global _janitor_task

    _history_event_handler = HistoryEventHandler()
    core_events.register_handler(_history_event_handler)

    _sampling_task = asyncio.create_task(sampling_task())
    _janitor_task = asyncio.create_task(janitor_task())

    await persist.ensure_index(PERSIST_COLLECTION, 'ts')


async def cleanup() -> None:
    if _sampling_task:
        _sampling_task.cancel()
        await _sampling_task

    if _janitor_task:
        _janitor_task.cancel()
        await _janitor_task




import asyncio
import logging
import time

from typing import Dict, Optional, Tuple


_CUSTOM_CLEANUP_INTERVAL = 5

logger = logging.getLogger(__name__)

_custom_dns_mapping: Dict[str, str] = {}
_custom_dns_mapping_timeouts: Dict[str, Tuple[int, int]] = {}
_custom_dns_cleanup_task: Optional[asyncio.Task] = None


def get_custom_dns_mapping_dict() -> Dict[str, str]:
    return _custom_dns_mapping


def set_custom_dns_mapping(hostname: str, ip_address: str, timeout: Optional[int] = None) -> None:
    logger.debug('setting custom DNS mapping %s -> %s with timeout %s', hostname, ip_address, timeout or 'none')

    _custom_dns_mapping[hostname] = ip_address
    if timeout:
        _custom_dns_mapping_timeouts[hostname] = (timeout, int(time.time()))


async def _custom_cleanup_loop() -> None:
    try:
        now = time.time()
        while True:
            for hostname, (timeout, added_time) in list(_custom_dns_mapping_timeouts.items()):
                if now - added_time > timeout:
                    logger.debug('custom DNS mapping for %s expired', hostname)
                    _custom_dns_mapping_timeouts.pop(hostname)
                    _custom_dns_mapping.pop(hostname, None)

            await asyncio.sleep(_CUSTOM_CLEANUP_INTERVAL)

    except asyncio.CancelledError:
        logger.debug('custom DNS mapping cleanup task cancelled')


async def init() -> None:
    global _custom_dns_cleanup_task

    _custom_dns_cleanup_task = asyncio.create_task(_custom_cleanup_loop())


async def cleanup() -> None:
    global _custom_dns_cleanup_task

    if _custom_dns_cleanup_task:
        _custom_dns_cleanup_task.cancel()
        await _custom_dns_cleanup_task
        _custom_dns_cleanup_task = None




from .json import JSONDriver

try:
    from .mongo import MongoDriver
except ImportError:
    pass

try:
    from .postgresql import PostgreSQLDriver
except ImportError:
    pass

try:
    from .redis import RedisDriver
except ImportError:
    pass




import logging

from qtoggleserver import persist
from qtoggleserver.core import api as core_api
from qtoggleserver.core import events as core_events
from qtoggleserver.core.typing import GenericJSONDict, GenericJSONList

from ..events import DashboardUpdateEvent


logger = logging.getLogger(__name__)


@core_api.api_call(core_api.ACCESS_LEVEL_VIEWONLY)
async def get_panels(request: core_api.APIRequest) -> GenericJSONList:
    return await persist.get_value('dashboard_panels', default=[])


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
async def put_panels(request: core_api.APIRequest, params: GenericJSONList) -> None:
    # core_api.validate(panels, PANELS_SCHEMA)  TODO: validate against schema

    await persist.set_value('dashboard_panels', params)

    await core_events.trigger(DashboardUpdateEvent(request=request, panels=params))


@core_api.api_call(core_api.ACCESS_LEVEL_VIEWONLY)
async def get_prefs(request: core_api.APIRequest) -> GenericJSONDict:
    prefs = await persist.get('frontend_prefs', id_=request.username) or {}
    prefs.pop('id', None)

    return prefs


@core_api.api_call(core_api.ACCESS_LEVEL_VIEWONLY)
async def put_prefs(request: core_api.APIRequest, params: GenericJSONDict) -> None:
    await persist.replace('frontend_prefs', id_=request.username, record=params)


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
async def get_frontend(request: core_api.APIRequest) -> GenericJSONDict:
    return {
        'prefs': sorted(await persist.query('frontend_prefs'), key=lambda p: p['id']),
        'dashboard_panels': await persist.get_value('dashboard_panels')
    }


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
async def put_frontend(request: core_api.APIRequest, params: GenericJSONDict) -> None:
    # core_api.validate(panels, FRONTEND_SCHEMA)  TODO: validate against schema

    logger.debug('restoring frontend configuration')

    await persist.remove('frontend_prefs')

    prefs = params.get('prefs')
    if prefs:
        for p in prefs:
            logger.debug('restoring frontend prefs for username "%s"', p.get('id'))
            await persist.insert('frontend_prefs', p)

    dashboard_panels = params.get('dashboard_panels', [])
    logger.debug('restoring dashboard panels')
    await persist.set_value('dashboard_panels', dashboard_panels)

    await core_events.trigger(DashboardUpdateEvent(request=request, panels=dashboard_panels))




import re

from typing import Optional

from qtoggleserver.core.typing import PortValue as CorePortValue

from .base import Expression, Evaluated
from .exceptions import UnexpectedCharacter, EmptyExpression


class LiteralValue(Expression):
    def __init__(self, value: CorePortValue, sexpression: str) -> None:
        self.value: CorePortValue = value
        self.sexpression: str = sexpression

    def __str__(self) -> str:
        return self.sexpression

    async def eval(self) -> Evaluated:
        return float(self.value)

    @staticmethod
    def parse(self_port_id: Optional[str], sexpression: str, pos: int) -> Expression:
        while sexpression and sexpression[0].isspace():
            sexpression = sexpression[1:]
            pos += 1

        while sexpression and sexpression[-1].isspace():
            sexpression = sexpression[:-1]

        if not sexpression:
            raise EmptyExpression()

        if sexpression == 'true':
            value = 1

        elif sexpression == 'false':
            value = 0

        else:
            try:
                value = int(sexpression)

            except ValueError:
                try:
                    value = float(sexpression)

                except ValueError:
                    m = re.match(r'-?\d+(\.?\d+)?', sexpression)
                    if m:
                        raise UnexpectedCharacter(sexpression[m.end()], pos + m.end()) from None

                    else:
                        raise UnexpectedCharacter(sexpression[0], pos) from None

        return LiteralValue(value, sexpression)




import logging

from qtoggleserver.core import api as core_api
from qtoggleserver.core.api import schema as core_api_schema
from qtoggleserver.core.typing import GenericJSONDict
from qtoggleserver.system import fwupdate


logger = logging.getLogger(__name__)


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
async def get_firmware(request: core_api.APIRequest) -> GenericJSONDict:
    current_version = await fwupdate.get_current_version()
    status = await fwupdate.get_status()

    if status == fwupdate.STATUS_IDLE:
        try:
            latest_version, latest_date, latest_url = await fwupdate.get_latest()

            return {
                'version': current_version,
                'latest_version': latest_version,
                'latest_date': latest_date,
                'latest_url': latest_url,
                'status': status
            }

        except Exception as e:
            logger.error('get latest firmware failed: %s', e, exc_info=True)

            return {
                'version': current_version,
                'status': status
            }

    else:
        return {
            'version': current_version,
            'status': status
        }


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
async def patch_firmware(request: core_api.APIRequest, params: GenericJSONDict) -> None:
    core_api_schema.validate(params, core_api_schema.PATCH_FIRMWARE)

    status = await fwupdate.get_status()
    if status not in (fwupdate.STATUS_IDLE, fwupdate.STATUS_ERROR):
        raise core_api.APIError(503, 'busy')

    if params.get('url'):
        await fwupdate.update_to_url(params['url'])

    else:  # Assuming params['version']
        await fwupdate.update_to_version(params['version'])




from collections import OrderedDict

from qtoggleserver.conf import settings
from qtoggleserver.core import api as core_api
from qtoggleserver.core.typing import GenericJSONDict
from qtoggleserver.system import conf as system_conf
from qtoggleserver.utils import conf as conf_utils


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
async def get_system(request: core_api.APIRequest) -> GenericJSONDict:
    return conf_utils.config_from_file(settings.source)


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
async def put_system(request: core_api.APIRequest, params: GenericJSONDict) -> None:
    # Update in-memory settings (no effect guaranteed without a restart, though)
    conf_utils.update_obj_from_dict(settings, OrderedDict(params))

    # Update config file
    system_conf.conf_file_from_dict(params)




import logging

from typing import Optional

import psutil

from qtoggleserver.conf import settings
from qtoggleserver.utils.cmd import run_get_cmd


_has_sensors_battery = None

logger = logging.getLogger(__name__)


class BatteryError(Exception):
    pass


def has_battery_support() -> bool:
    global _has_sensors_battery

    if _has_sensors_battery is None:
        try:
            _ = psutil.sensors_battery().percent
            _has_sensors_battery = True
            logger.debug('battery sensor support detected')

        except AttributeError:
            _has_sensors_battery = False

    return _has_sensors_battery or bool(settings.system.battery.get_cmd)


def get_battery_level() -> Optional[int]:
    if settings.system.battery.get_cmd:
        return int(run_get_cmd(
            settings.system.battery.get_cmd,
            cmd_name='battery',
            exc_class=BatteryError,
            required_fields=['level']
        )['level'])

    if _has_sensors_battery:
        return int(psutil.sensors_battery().percent)




import logging

from qtoggleserver.conf import settings
from qtoggleserver.core import api as core_api
from qtoggleserver.core.typing import GenericJSONList
from qtoggleserver.system import conf as system_conf


logger = logging.getLogger(__name__)


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
async def get_backup_endpoints(request: core_api.APIRequest) -> GenericJSONList:
    endpoints = []

    if system_conf.can_write_conf_file():
        endpoints.append({
            'path': '/system',
            'display_name': 'System Configuration',
            'restore_method': 'PUT',
            'order': 5
        })

    if settings.frontend.enabled:
        endpoints.append({
            'path': '/frontend',
            'display_name': 'App Configuration',
            'restore_method': 'PUT',
            'order': 45
        })

    return endpoints




import asyncio
import logging
import ssl

from typing import List, Optional

from tornado.web import Application, HTTPServer, RequestHandler, URLSpec
from qui.web import tornado as qui_tornado

from qtoggleserver import system
from qtoggleserver.conf import settings
from qtoggleserver.core import history
from qtoggleserver.slaves.discover import is_enabled as is_discover_enabled
from qtoggleserver.web import handlers


logger = logging.getLogger(__name__)

_application: Optional[Application] = None
_server: Optional[HTTPServer] = None


def _log_request(handler: RequestHandler) -> None:
    if handler.get_status() < 400:
        log_method = logger.debug

    elif handler.get_status() < 500:
        log_method = logger.warning

    else:
        log_method = logger.error

    request_time = 1000.0 * handler.request.request_time()
    log_method('%d %s %.2fms', handler.get_status(), handler._request_summary(), request_time)


def _make_routing_table() -> List[URLSpec]:
    handlers_list = []

    # Frontend

    if settings.frontend.enabled:
        handlers_list += qui_tornado.make_routing_table()

        handlers_list += [
            URLSpec(r'^/api/frontend/dashboard/panels/?$', handlers.DashboardPanelsHandler),
            URLSpec(r'^/api/frontend/prefs/?$', handlers.PrefsHandler),
            URLSpec(r'^/api/frontend/?$', handlers.FrontendHandler),
        ]

    handlers_list += [
        # Device management
        URLSpec(r'^/api/device/?$', handlers.DeviceHandler),
        URLSpec(r'^/api/reset/?$', handlers.ResetHandler),
        URLSpec(r'^/api/access/?$', handlers.AccessHandler),

        # Port management
        URLSpec(r'^/api/ports/?$', handlers.PortsHandler),
        URLSpec(r'^/api/ports/(?P<port_id>[A-Za-z0-9_.-]+)/?$', handlers.PortHandler),

        # Port values
        URLSpec(r'^/api/ports/(?P<port_id>[A-Za-z0-9_.-]+)/value/?$', handlers.PortValueHandler),
    ]

    if settings.core.sequences_support:
        handlers_list += [
            URLSpec(r'^/api/ports/(?P<port_id>[A-Za-z0-9_.-]+)/sequence/?$', handlers.PortSequenceHandler)
        ]

    if history.is_enabled():
        handlers_list += [
            URLSpec(r'^/api/ports/(?P<port_id>[A-Za-z0-9_.-]+)/history/?$', handlers.PortHistoryHandler),
        ]

    if settings.core.backup_support:
        handlers_list += [
            URLSpec(r'^/api/backup/endpoints/?$', handlers.BackupEndpointsHandler)
        ]

    # Firmware

    if settings.system.fwupdate.driver:
        handlers_list += [
            URLSpec(r'^/api/firmware/?$', handlers.FirmwareHandler)
        ]

    # Slave devices management

    if settings.slaves.enabled:
        handlers_list += [
            URLSpec(r'^/api/devices/?$', handlers.SlaveDevicesHandler),
            URLSpec(r'^/api/devices/(?P<name>[A-Za-z0-9_-]+)/?$', handlers.SlaveDeviceHandler),
            URLSpec(r'^/api/devices/(?P<name>[A-Za-z0-9_-]+)/events/?$', handlers.SlaveDeviceEventsHandler),
            URLSpec(r'^/api/devices/(?P<name>[A-Za-z0-9_-]+)/forward/(?P<path>.+)$', handlers.SlaveDeviceForwardHandler)
        ]

        if is_discover_enabled():
            handlers_list += [
                URLSpec(r'^/api/discovered/?$', handlers.DiscoveredHandler),
                URLSpec(r'^/api/discovered/(?P<name>[A-Za-z0-9_-]+)/?$', handlers.DiscoveredDeviceHandler)
            ]

    # Notifications

    if settings.webhooks.enabled:
        handlers_list += [
            URLSpec(r'^/api/webhooks/?$', handlers.WebhooksHandler)
        ]

    if settings.core.listen_support:
        handlers_list += [
            URLSpec(r'^/api/listen/?$', handlers.ListenHandler)
        ]

    # Reverse API calls

    if settings.reverse.enabled:
        handlers_list += [
            URLSpec(r'^/api/reverse/?$', handlers.ReverseHandler)
        ]

    # System API calls

    if system.conf.can_write_conf_file():
        handlers_list += [
            URLSpec(r'^/api/system/?$', handlers.SystemHandler)
        ]

    # Introspection API

    if settings.debug:
        handlers_list += [
            URLSpec(r'^/api/introspect/?$', handlers.IntrospectHandler)
        ]

    # Default 404 API handler

    handlers_list += [
        URLSpec(r'^/api/.*$', handlers.NoSuchFunctionHandler)
    ]

    return handlers_list


def get_application() -> Application:
    global _application

    if _application is None:
        _application = Application(
            handlers=_make_routing_table(),
            debug=False,
            compress_response=settings.server.compress_response,
            log_function=_log_request
        )

    return _application


def get_server() -> Optional[HTTPServer]:
    return _server


async def init() -> None:
    global _server

    address, port = settings.server.addr, settings.server.port

    ssl_context = None
    if settings.server.https.cert_file and settings.server.https.key_file:
        logger.info('setting up HTTPS using certificate from %s', settings.server.https.cert_file)
        ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        ssl_context.load_cert_chain(settings.server.https.cert_file, settings.server.https.key_file)

    app = get_application()

    try:
        _server = app.listen(port, address, ssl_options=ssl_context)
        logger.info('server listening on %s:%s', address, port)

    except Exception as e:
        logger.error('server listen failed: %s', e)
        raise


async def cleanup() -> None:
    global _server

    if not _server:
        return

    _server.stop()

    # Allow a small amount of time for web server callbacks to complete
    await asyncio.sleep(1)




import importlib
import inspect
import traceback

from typing import Dict

from qtoggleserver import slaves
from qtoggleserver import system
from qtoggleserver.conf import settings
from qtoggleserver.core import api as core_api
from qtoggleserver.core import device as core_device
from qtoggleserver.core import history as core_history
from qtoggleserver.core import main
from qtoggleserver.core import ports as core_ports
from qtoggleserver.core import reverse as core_reverse
from qtoggleserver.core import sessions as core_sessions
from qtoggleserver.core import vports as core_vports
from qtoggleserver.core import webhooks as core_webhooks
from qtoggleserver.core.api import schema as core_api_schema
from qtoggleserver.core.typing import GenericJSONDict, GenericJSONList


@core_api.api_call(core_api.ACCESS_LEVEL_NONE)
async def get_access(request: core_api.APIRequest) -> Dict[str, str]:
    return {
        'level': core_api.ACCESS_LEVEL_MAPPING[request.access_level]
    }


@core_api.api_call(core_api.ACCESS_LEVEL_VIEWONLY)
async def get_listen(
    request: core_api.APIRequest
) -> GenericJSONList:

    session_id = request.headers.get('Session-Id')
    if not session_id:
        raise core_api.APIError(400, 'missing-header', header='Session-Id')

    timeout = request.query.get('timeout')
    if timeout is not None:
        try:
            timeout = int(timeout)

        except ValueError:
            raise core_api.APIError(400, 'invalid-field', field='timeout') from None

        if timeout < 1 or timeout > 3600:
            raise core_api.APIError(400, 'invalid-field', field='timeout')

    else:
        timeout = 60  # default

    session = core_sessions.get(session_id)
    events = await session.reset_and_wait(timeout, request.access_level)

    return [await e.to_json() for e in events]


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
async def post_reset(request: core_api.APIRequest, params: GenericJSONDict) -> None:
    core_api_schema.validate(params, core_api_schema.POST_RESET)

    factory = params.get('factory')

    if factory:
        core_api.logger.info('resetting to factory defaults')

        await core_ports.reset()
        await core_vports.reset()
        await core_device.reset()
        if settings.webhooks.enabled:
            await core_webhooks.reset()
        if settings.reverse.enabled:
            await core_reverse.reset()
        if settings.slaves.enabled:
            await slaves.reset_ports()
            await slaves.reset_slaves()
        if core_history.is_enabled():
            await core_history.reset()

    main.loop.call_later(2, system.reboot)


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
async def post_introspect(request: core_api.APIRequest, params: GenericJSONDict) -> GenericJSONDict:
    core_api_schema.validate(params, core_api_schema.POST_INTROSPECT)

    exc_str = None
    res_str = None

    try:
        imports = params.get('imports', [])
        extra_locals = {}
        for imp in imports:
            extra_locals[imp.split('.')[0]] = importlib.__import__(imp)

        result = eval(params['code'], globals(), dict(locals(), **extra_locals))
        if inspect.isawaitable(result):
            result = await result

        res_str = str(result)

    except Exception:
        exc_str = traceback.format_exc()

    return {
        'result': res_str,
        'exception': exc_str
    }




from __future__ import annotations

import logging
import queue

from typing import Optional

from tornado.httpclient import AsyncHTTPClient, HTTPRequest, HTTPResponse

from qtoggleserver import persist
from qtoggleserver.conf import settings
from qtoggleserver.core import responses as core_responses
from qtoggleserver.core.api import auth as core_api_auth
from qtoggleserver.core.device import attrs as core_device_attrs
from qtoggleserver.core.typing import GenericJSONDict
from qtoggleserver.utils import json as json_utils


logger = logging.getLogger(__name__)

_webhooks: Optional[Webhooks] = None


# TODO: add password support


class WebhooksError(Exception):
    pass


class InvalidParamError(WebhooksError):
    def __init__(self, param: str) -> None:
        self.param: str = param

        super().__init__(f'Invalid field: {param}')


class WebhooksRequest:
    def __init__(self, body: GenericJSONDict) -> None:
        self.body: GenericJSONDict = body
        self.retries: int = 0


class Webhooks:
    def __init__(
        self,
        scheme: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[int] = None,
        path: Optional[str] = None,
        timeout: Optional[int] = None,
        retries: Optional[int] = None,
        **kwargs
    ) -> None:

        # The enabled value comes with kwargs but is ignored; webhooks will be explicitly enabled afterwards

        self._enabled: bool = False
        self._scheme: Optional[str] = scheme
        self._host: Optional[str] = host
        self._port: Optional[int] = port
        self._path: Optional[str] = path
        self._timeout: Optional[int] = timeout
        self._retries: Optional[int] = retries
        self._url: Optional[str] = None

        self._queue: queue.Queue = queue.Queue(settings.core.event_queue_size)

    def __str__(self) -> str:
        s = 'webhooks'

        if not self._enabled:
            s += ' [disabled]'

        if self._scheme:
            s += ' ' + self.get_url()

        return s

    def is_enabled(self) -> bool:
        return self._enabled

    def enable(self) -> None:
        if self._enabled:
            return

        logger.debug('enabling %s', self)

        self._enabled = True

    def disable(self) -> None:
        if not self._enabled:
            return

        logger.debug('disabling %s', self)

        self._enabled = False

        # Drop all queued requests
        while not self._queue.empty():
            self._queue.get()

    def get_url(self) -> str:
        if not self._url:
            if self._scheme == 'http' and self._port == 80 or self._scheme == 'https' and self._port == 443:
                self._url = f'{self._scheme}://{self._host}{self._path}'

            else:
                self._url = f'{self._scheme}://{self._host}:{self._port}{self._path}'

        return self._url

    def call(self, body: GenericJSONDict) -> None:
        if not self._enabled:
            return

        request = WebhooksRequest(body)

        if self._queue.empty():
            return self._request(request)

        try:
            self._queue.put(request)

        except queue.Full:
            logger.error('%s: queue is full', self)
            return

    def _request(self, request: WebhooksRequest) -> None:
        if not self._enabled:
            return

        url = self.get_url()
        body = json_utils.dumps(request.body)

        def on_response(response: HTTPResponse) -> None:
            try:
                core_responses.parse(response)

            except core_responses.Error as e:
                logger.error('%s: call failed: %s', self, e)

                # Retry mechanism
                if not self._retries:
                    return self._check_pending()

                if request.retries <= self._retries:
                    request.retries += 1
                    logger.debug('%s: resending request (retry %s/%s)', self, request.retries, self._retries)

                    self._request(request)

                else:
                    self._check_pending()

            else:
                logger.debug('%s: call succeeded', self)
                self._check_pending()

        http_client = AsyncHTTPClient()
        # TODO use webhooks password
        headers = {
            'Content-Type': json_utils.JSON_CONTENT_TYPE,
            'Authorization': core_api_auth.make_auth_header(
                core_api_auth.ORIGIN_DEVICE,
                username=None,
                password_hash=core_device_attrs.normal_password_hash
            )
        }
        request = HTTPRequest(
            url,
            'POST',
            headers=headers,
            body=body,
            connect_timeout=self._timeout,
            request_timeout=self._timeout,
            follow_redirects=True
        )

        logger.debug('%s: calling', self)

        http_client.fetch(request, callback=on_response)  # TODO implement me using await

    def _check_pending(self) -> None:
        try:
            request = self._queue.get(block=False)

        except queue.Empty:
            return

        self._request(request)

    def to_json(self) -> GenericJSONDict:
        d = {
            'enabled': self._enabled,
            'scheme': self._scheme,
            'host': self._host,
            'port': self._port,
            'path': self._path,
            'timeout': self._timeout,
            'retries': self._retries
        }

        return d


def get() -> Optional[Webhooks]:
    return _webhooks


def setup(
    enabled: bool,
    scheme: Optional[str] = None,
    host: Optional[str] = None,
    port: Optional[int] = None,
    path: Optional[str] = None,
    timeout: Optional[int] = None,
    retries: Optional[int] = None,
    **kwargs
) -> None:

    global _webhooks

    if _webhooks and _webhooks.is_enabled():
        _webhooks.disable()

    _webhooks = Webhooks(scheme, host, port, path, timeout, retries)
    if enabled:
        _webhooks.enable()


async def load() -> None:
    data = await persist.get_value('webhooks')
    if data is None:
        setup(enabled=False)
        logger.debug('loaded %s', _webhooks)

        return

    setup(**data)
    logger.debug('loaded %s', _webhooks)


async def save() -> None:
    if _webhooks is None:
        return

    logger.debug('saving persisted data')
    await persist.set_value('webhooks', _webhooks.to_json())


async def reset() -> None:
    logger.debug('clearing persisted data')
    await persist.remove('webhooks')


async def init() -> None:
    logger.debug('loading persisted data')
    await load()


async def cleanup() -> None:
    pass




import logging.handlers

from typing import Optional


class FifoMemoryHandler(logging.handlers.MemoryHandler):
    def __init__(self, capacity: int) -> None:
        super().__init__(capacity, flushLevel=logging.FATAL + 1)

    def flush(self) -> None:
        self.acquire()

        while len(self.buffer) > self.capacity:
            self.buffer.pop(0)

        self.release()


class LoggableMixin:
    def __init__(self, name: Optional[str], parent_logger: logging.Logger) -> None:
        if name:
            name = f'{parent_logger.name}.{name}'

        else:
            name = parent_logger.name

        self._parent_logger: logging.Logger = parent_logger
        self._logger: logging.Logger = logging.getLogger(name)

    def log(self, level: int, msg: str, *args, **kwargs) -> None:
        self._logger.log(level, msg, *args, **kwargs)

    def debug(self, msg: str, *args, **kwargs) -> None:
        self.log(logging.DEBUG, msg, *args, **kwargs)

    def info(self, msg: str, *args, **kwargs) -> None:
        self.log(logging.INFO, msg, *args, **kwargs)

    def warning(self, msg: str, *args, **kwargs) -> None:
        self.log(logging.WARNING, msg, *args, **kwargs)

    def error(self, msg: str, *args, **kwargs) -> None:
        self.log(logging.ERROR, msg, *args, **kwargs)

    def set_logger_name(self, name: str) -> None:
        if name:
            self._logger = logging.getLogger(f'{self._parent_logger.name}.{name}')

        else:
            self._logger = self._parent_logger




import abc

from typing import Any, Dict, Iterable, List, Optional, Tuple

from .typing import Id, Record


class BaseDriver(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    async def query(
        self,
        collection: str,
        fields: Optional[List[str]],
        filt: Dict[str, Any],
        sort: List[Tuple[str, bool]],
        limit: Optional[int]
    ) -> Iterable[Record]:
        return []

    @abc.abstractmethod
    async def insert(self, collection: str, record: Record) -> Id:
        return '1'  # Returns the inserted record id

    @abc.abstractmethod
    async def update(self, collection: str, record_part: Record, filt: Dict[str, Any]) -> int:
        return 0  # Returns the number of updated records

    @abc.abstractmethod
    async def replace(self, collection: str, id_: Id, record: Record) -> bool:
        return False  # Returns True if matched and replaced

    @abc.abstractmethod
    async def remove(self, collection: str, filt: Dict[str, Any]) -> int:
        return 0  # Returns the number of removed records

    async def ensure_index(self, collection: str, index: List[Tuple[str, bool]]) -> None:
        pass

    async def cleanup(self) -> None:
        pass

    def is_history_supported(self) -> bool:
        return False




from qtoggleserver.core.typing import PortValue as CorePortValue
from qtoggleserver.core.typing import GenericJSONDict


class ExpressionException(Exception):
    pass


class ExpressionParseError(ExpressionException):
    def to_json(self) -> GenericJSONDict:
        return {
            'reason': 'parse-error'
        }


class UnknownFunction(ExpressionParseError):
    def __init__(self, name: str, pos: int) -> None:
        self.name: str = name
        self.pos: int = pos

        super().__init__(f'Unknown function "{name}"')

    def to_json(self) -> GenericJSONDict:
        return {
            'reason': 'unknown-function',
            'token': self.name,
            'pos': self.pos
        }


class InvalidNumberOfArguments(ExpressionParseError):
    def __init__(self, name: str, pos: int) -> None:
        self.name: str = name
        self.pos: int = pos

        super().__init__(f'Invalid number of arguments for function "{name}"')

    def to_json(self) -> GenericJSONDict:
        return {
            'reason': 'invalid-number-of-arguments',
            'token': self.name,
            'pos': self.pos
        }


class InvalidArgumentKind(ExpressionParseError):
    def __init__(self, name: str, pos: int, num: int) -> None:
        self.name: str = name
        self.pos: int = pos
        self.num: int = num

        super().__init__(f'Invalid argument {num} kind for "{name}" at position {pos}')

    def to_json(self) -> GenericJSONDict:
        return {
            'reason': 'invalid-argument-kind',
            'token': self.name,
            'pos': self.pos,
            'num': self.num
        }


class UnbalancedParentheses(ExpressionParseError):
    def __init__(self, pos: int) -> None:
        self.pos: int = pos

        super().__init__('Unbalanced parentheses')

    def to_json(self) -> GenericJSONDict:
        return {
            'reason': 'unbalanced-parentheses',
            'pos': self.pos
        }


class UnexpectedEnd(ExpressionParseError):
    def __init__(self) -> None:
        super().__init__('Expression is unterminated')

    def to_json(self) -> GenericJSONDict:
        return {
            'reason': 'unexpected-end'
        }


class CircularDependency(ExpressionParseError):
    def __init__(self, port_id: str) -> None:
        super().__init__(f'Expression creates a dependency loop via port "{port_id}"')

    def to_json(self) -> GenericJSONDict:
        return {
            'reason': 'circular-dependency'
        }


class ExternalDependency(ExpressionParseError):
    def __init__(self, port_id: str, pos: int) -> None:
        self.port_id = port_id
        self.pos = pos

        super().__init__(f'External dependency on port "{port_id}"')

    def to_json(self) -> GenericJSONDict:
        return {
            'reason': 'external-dependency',
            'token': self.port_id,
            'pos': self.pos
        }


class UnexpectedCharacter(ExpressionParseError):
    def __init__(self, c: str, pos: int) -> None:
        self.c = c
        self.pos: int = pos

        super().__init__(f'Unexpected character "{c}" at position {pos}')

    def to_json(self) -> GenericJSONDict:
        return {
            'reason': 'unexpected-character',
            'token': self.c,
            'pos': self.pos
        }


class EmptyExpression(ExpressionParseError):
    def __init__(self) -> None:
        super().__init__('Expression is empty')

    def to_json(self) -> GenericJSONDict:
        return {
            'reason': 'empty-expression'
        }


class ExpressionEvalError(ExpressionException):
    pass


class InvalidArgumentValue(ExpressionEvalError):
    def __init__(self, arg_no: int, value: CorePortValue) -> None:
        self.arg_no: int = arg_no
        self.value: CorePortValue = value

        super().__init__(f'Invalid argument {self.arg_no}: {self.value}')


class UnknownPortId(ExpressionEvalError):
    def __init__(self, port_id: str) -> None:
        self.port_id = port_id

        super().__init__(f'Unknown port "{port_id}"')


class DisabledPort(ExpressionEvalError):
    def __init__(self, port_id: str) -> None:
        self.port_id = port_id

        super().__init__(f'Port "{port_id}" is disabled')


class PortValueUnavailable(ExpressionEvalError):
    def __init__(self, port_id: str) -> None:
        self.port_id = port_id

        super().__init__(f'Port "{port_id}" value is undefined')


class ExpressionArithmeticError(ExpressionEvalError):
    def __init__(self) -> None:
        super().__init__('Expression arithmetic error')


class EvalSkipped(ExpressionEvalError):
    def __init__(self) -> None:
        super().__init__('Evaluation skipped')




from .base import Evaluated
from .exceptions import ExpressionArithmeticError
from .functions import function, Function


@function('ADD')
class AddFunction(Function):
    MIN_ARGS = 2

    async def eval(self) -> Evaluated:
        return sum(await self.eval_args())


@function('SUB')
class SubFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    async def eval(self) -> Evaluated:
        eval_args = await self.eval_args()
        return eval_args[0] - eval_args[1]


@function('MUL')
class MulFunction(Function):
    MIN_ARGS = 2

    async def eval(self) -> Evaluated:
        r = 1
        for e in await self.eval_args():
            r *= e

        return r


@function('DIV')
class DivFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    async def eval(self) -> Evaluated:
        eval_args = await self.eval_args()

        if eval_args[1]:
            return eval_args[0] / eval_args[1]

        else:
            raise ExpressionArithmeticError


@function('MOD')
class ModFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    async def eval(self) -> Evaluated:
        eval_args = await self.eval_args()

        if eval_args[1]:
            return eval_args[0] % eval_args[1]

        else:
            raise ExpressionArithmeticError


@function('POW')
class PowFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    async def eval(self) -> Evaluated:
        eval_args = await self.eval_args()

        return eval_args[0] ** eval_args[1]




import asyncio
import inspect
import time

from typing import Any, Callable, List

from qtoggleserver import slaves
from qtoggleserver.conf import settings
from qtoggleserver.core import api as core_api
from qtoggleserver.core import events as core_events
from qtoggleserver.core import history as core_history
from qtoggleserver.core import main as core_main
from qtoggleserver.core import ports as core_ports
from qtoggleserver.core import vports as core_vports
from qtoggleserver.core.api import schema as core_api_schema
from qtoggleserver.core.typing import Attribute, Attributes, GenericJSONDict, GenericJSONList, NullablePortValue
from qtoggleserver.core.typing import PortValue
from qtoggleserver.slaves import ports as slaves_ports
from qtoggleserver.slaves import devices as slaves_devices
from qtoggleserver.utils import json as json_utils


async def add_virtual_port(attrs: GenericJSONDict) -> core_ports.BasePort:
    id_ = attrs['id']
    type_ = attrs['type']
    min_ = attrs.get('min')
    max_ = attrs.get('max')
    integer = attrs.get('integer')
    step = attrs.get('step')
    choices = attrs.get('choices')

    core_api.logger.debug('adding port "%s"', id_)

    if core_ports.get(id_):
        raise core_api.APIError(400, 'duplicate-port')

    if len(core_vports.all_port_args()) >= settings.core.virtual_ports:
        raise core_api.APIError(400, 'too-many-ports')

    await core_vports.add(id_, type_, min_, max_, integer, step, choices)
    port = await core_ports.load_one(
        'qtoggleserver.core.vports.VirtualPort',
        {
            'id_': id_,
            'type_': type_,
            'min_': min_,
            'max_': max_,
            'integer': integer,
            'step': step,
            'choices': choices
        }
    )

    # A virtual port is enabled by default
    await port.enable()
    await port.save()

    return port


async def set_port_attrs(port: core_ports.BasePort, attrs: GenericJSONDict, ignore_extra_attrs: bool) -> None:
    non_modifiable_attrs = await port.get_non_modifiable_attrs()

    def unexpected_field_code(field: str) -> str:
        if field in non_modifiable_attrs:
            return 'attribute-not-modifiable'

        else:
            return 'no-such-attribute'

    schema = await port.get_schema()
    if ignore_extra_attrs:
        schema = dict(schema)
        schema['additionalProperties'] = True  # Ignore non-existent and non-modifiable attributes

    core_api_schema.validate(
        attrs,
        schema,
        unexpected_field_code=unexpected_field_code,
        unexpected_field_name='attribute'
    )

    # Step validation
    attrdefs = await port.get_attrdefs()
    for name, value in attrs.items():
        attrdef = attrdefs.get(name)
        if attrdef is None:
            continue

        step = attrdef.get('step')
        min_ = attrdef.get('min')
        if None not in (step, min_) and step != 0 and (value - min_) % step:
            raise core_api.APIError(400, 'invalid-field', field=name)

    errors_by_name = {}

    async def set_attr(attr_name: str, attr_value: Attribute) -> None:
        core_api.logger.debug('setting attribute %s = %s on %s', attr_name, json_utils.dumps(attr_value), port)

        try:
            await port.set_attr(attr_name, attr_value)

        except Exception as e1:
            errors_by_name[attr_name] = e1

    value = attrs.pop('value', None)

    if attrs:
        await asyncio.wait([set_attr(n, v) for n, v in attrs.items()])

    if errors_by_name:
        name, error = next(iter(errors_by_name.items()))

        if isinstance(error, core_api.APIError):
            raise error

        elif isinstance(error, core_ports.InvalidAttributeValue):
            raise core_api.APIError(400, 'invalid-field', field=name, details=error.details)

        elif isinstance(error, core_ports.PortTimeout):
            raise core_api.APIError(504, 'port-timeout')

        elif isinstance(error, core_ports.PortError):
            raise core_api.APIError(502, 'port-error', code=str(error))

        else:
            # Transform any unhandled exception into APIError(500)
            raise core_api.APIError(500, 'unexpected-error', message=str(error)) from error

    # If value is supplied among attrs, use it to update port value, but in background and ignoring any errors
    if value is not None and port.is_enabled():
        asyncio.create_task(port.write_transformed_value(value, reason=core_ports.CHANGE_REASON_API))

    await port.save()


async def wrap_error_with_port_id(port_id: str, func: Callable, *args, **kwargs) -> Any:
    try:
        result = func(*args, **kwargs)
        if inspect.isawaitable(result):
            result = await result

    except core_api.APIError as e:
        raise core_api.APIError(
            status=e.status,
            code=e.code,
            id=port_id,
            **e.params
        )

    except Exception as e:
        raise core_api.APIError(
            status=500,
            code='unexpected-error',
            message=str(e),
            id=port_id
        )

    return result


@core_api.api_call(core_api.ACCESS_LEVEL_VIEWONLY)
async def get_ports(request: core_api.APIRequest) -> List[Attributes]:
    return [await port.to_json() for port in sorted(core_ports.get_all(), key=lambda p: p.get_id())]


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
async def put_ports(request: core_api.APIRequest, params: GenericJSONList) -> None:
    if not settings.core.backup_support:
        raise core_api.APIError(404, 'no-such-function')

    core_api_schema.validate(
        params,
        core_api_schema.PUT_PORTS
    )

    core_api.logger.debug('restoring ports')

    # Disable event handling during the processing of this request, as we're going to trigger a full-update at the end
    core_events.disable()

    # Temporarily disable core updating (port polling, expression evaluating and value-change handling)
    core_main.disable_updating()

    try:
        # Remove all (local) virtual ports
        for port in core_ports.get_all():
            if not isinstance(port, core_vports.VirtualPort):
                continue

            await port.remove()
            await core_vports.remove(port.get_id())

        # Reset ports
        await core_ports.reset()
        if settings.slaves.enabled:
            await slaves.reset_ports()
        for port in core_ports.get_all():
            await port.reset()

        add_port_schema = dict(core_api_schema.POST_PORTS)
        add_port_schema['additionalProperties'] = True

        # Restore supplied attributes
        for attrs in params:
            id_ = attrs.get('id')
            if id_ is None:
                core_api.logger.warning('ignoring entry without id')
                continue

            port = core_ports.get(id_)

            # Virtual ports must be added first (unless they belong to a slave)
            virtual = attrs.get('virtual')
            if port is not None:  # Port already exists so it probably belongs to a slave
                virtual = False
            for slave in slaves_devices.get_all():
                if id_.startswith(f'{slave.get_name()}.'):  # id indicates that port belongs to a slave
                    virtual = False
                    break
            if 'provisioning' in attrs:  # A clear indication that port belongs to a slave
                virtual = False

            if virtual:
                await wrap_error_with_port_id(
                    id_,
                    core_api_schema.validate,
                    attrs,
                    add_port_schema
                )
                port = await wrap_error_with_port_id(
                    id_,
                    add_virtual_port,
                    attrs
                )

            if port is None:
                core_api.logger.warning('ignoring unknown port id "%s"', id_)
                continue

            if isinstance(port, slaves_ports.SlavePort):
                core_api.logger.debug('restoring slave port "%s"', id_)

                # For slave ports, ignore any attributes that are not kept on master
                attrs = {n: v for n, v in attrs.items() if n in ('tag', 'expression', 'expires')}

            else:
                core_api.logger.debug('restoring local port "%s"', id_)

            await wrap_error_with_port_id(
                id_,
                set_port_attrs,
                port,
                attrs,
                ignore_extra_attrs=True
            )

    finally:
        core_main.enable_updating()
        core_events.enable()

    await core_events.trigger_full_update()

    core_api.logger.debug('ports restore done')


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
async def patch_port(request: core_api.APIRequest, port_id: str, params: Attributes) -> None:
    port = core_ports.get(port_id)
    if port is None:
        raise core_api.APIError(404, 'no-such-port')

    await set_port_attrs(port, params, ignore_extra_attrs=False)


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
async def post_ports(request: core_api.APIRequest, params: GenericJSONDict) -> Attributes:
    core_api_schema.validate(params, core_api_schema.POST_PORTS)
    port = await add_virtual_port(params)

    return await port.to_json()


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
async def delete_port(request: core_api.APIRequest, port_id: str) -> None:
    port = core_ports.get(port_id)
    if not port:
        raise core_api.APIError(404, 'no-such-port')

    if not isinstance(port, core_vports.VirtualPort):
        raise core_api.APIError(400, 'port-not-removable')

    await port.remove()
    await core_vports.remove(port_id)


@core_api.api_call(core_api.ACCESS_LEVEL_VIEWONLY)
async def get_port_value(request: core_api.APIRequest, port_id: str) -> NullablePortValue:
    port = core_ports.get(port_id)
    if port is None:
        raise core_api.APIError(404, 'no-such-port')

    if not port.is_enabled():
        return

    # TODO
    # Given the fact that get_last_read_value() simply returns last cached read value, and the fact that API specs
    # indicate that 502/504 be returned by GET /port/[id]/value in case of errors, we should remember the last error
    # generated by read_value() and return it here, if any.

    return port.get_last_read_value()


@core_api.api_call(core_api.ACCESS_LEVEL_NORMAL)
async def patch_port_value(request: core_api.APIRequest, port_id: str, params: PortValue) -> None:
    port = core_ports.get(port_id)
    if port is None:
        raise core_api.APIError(404, 'no-such-port')

    try:
        core_api_schema.validate(params, await port.get_value_schema())

    except core_api.APIError:
        # Transform any validation error into an invalid-field APIError for value
        raise core_api.APIError(400, 'invalid-value') from None

    value = params

    # Step validation
    step = await port.get_attr('step')
    min_ = await port.get_attr('min')
    if None not in (step, min_) and step != 0 and (value - min_) % step:
        raise core_api.APIError(400, 'invalid-value')

    if not port.is_enabled():
        raise core_api.APIError(400, 'port-disabled')

    if not await port.is_writable():
        raise core_api.APIError(400, 'read-only-port')

    old_value = port.get_last_read_value()

    try:
        await port.write_transformed_value(value, reason=core_ports.CHANGE_REASON_API)

    except core_ports.PortTimeout as e:
        raise core_api.APIError(504, 'port-timeout') from e

    except core_ports.PortError as e:
        raise core_api.APIError(502, 'port-error', code=str(e)) from e

    except core_api.APIError:
        raise

    except Exception as e:
        # Transform any unhandled exception into APIError(500)
        raise core_api.APIError(500, 'unexpected-error', message=str(e)) from e

    # If port value hasn't really changed, use 202 Accepted to inform consumer that new value hasn't been applied yet
    current_value = port.get_last_read_value()
    if (old_value == current_value) and (old_value != value):
        port.debug("API supplied value hasn't been applied right away")
        raise core_api.APIAccepted()


@core_api.api_call(core_api.ACCESS_LEVEL_NORMAL)
async def patch_port_sequence(request: core_api.APIRequest, port_id: str, params: GenericJSONDict) -> None:
    port = core_ports.get(port_id)
    if port is None:
        raise core_api.APIError(404, 'no-such-port')

    core_api_schema.validate(params, core_api_schema.PATCH_PORT_SEQUENCE)

    values = params['values']
    delays = params['delays']
    repeat = params['repeat']

    if len(values) != len(delays):
        raise core_api.APIError(400, 'invalid-field', field='delays')

    value_schema = await port.get_value_schema()
    step = await port.get_attr('step')
    min_ = await port.get_attr('min')
    for value in values:
        # Translate any APIError generated when validating value schema into an invalid-field APIError on value
        try:
            core_api_schema.validate(value, value_schema)

        except core_api.APIError:
            raise core_api.APIError(400, 'invalid-field', field='values') from None

        # Step validation
        if None not in (step, min_) and step != 0 and (value - min_) % step:
            raise core_api.APIError(400, 'invalid-field', field='values')

    if not port.is_enabled():
        raise core_api.APIError(400, 'port-disabled')

    if not await port.is_writable():
        raise core_api.APIError(400, 'read-only-port')

    if await port.get_attr('expression'):
        raise core_api.APIError(400, 'port-with-expression')

    try:
        await port.set_sequence(values, delays, repeat)

    except Exception as e:
        # Transform any unhandled exception into APIError(500)
        raise core_api.APIError(500, 'unexpected-error', message=str(e)) from e


@core_api.api_call(core_api.ACCESS_LEVEL_VIEWONLY)
async def get_port_history(request: core_api.APIRequest, port_id: str) -> GenericJSONList:
    port = core_ports.get(port_id)
    if port is None:
        raise core_api.APIError(404, 'no-such-port')

    query = request.query

    from_str = query.get('from')
    timestamps_str = query.get('timestamps')

    if from_str is None and timestamps_str is None:
        raise core_api.APIError(400, 'missing-field', field='from')

    if from_str:
        try:
            from_timestamp = int(from_str)

        except ValueError:
            raise core_api.APIError(400, 'invalid-field', field='from')

        if from_timestamp < 0:
            raise core_api.APIError(400, 'invalid-field', field='from')

    else:
        from_timestamp = None

    to_str = query.get('to')
    to_timestamp = int(time.time() * 1000)
    if to_str is not None:
        try:
            to_timestamp = int(to_str)

        except ValueError:
            raise core_api.APIError(400, 'invalid-field', field='to') from None

        if to_timestamp < 0:
            raise core_api.APIError(400, 'invalid-field', field='to')

    limit_str = query.get('limit')
    limit = 1000  # default
    if limit_str is not None:
        try:
            limit = int(limit_str)

        except ValueError:
            raise core_api.APIError(400, 'invalid-field', field='limit') from None

        if limit < 1 or limit > 10000:
            raise core_api.APIError(400, 'invalid-field', field='limit')

    timestamps = None
    if timestamps_str is not None:
        timestamps = timestamps_str.split(',')
        try:
            timestamps = [int(t) for t in timestamps]

        except ValueError:
            raise core_api.APIError(400, 'invalid-field', field='timestamps')

        if any((t < 0) for t in timestamps):
            raise core_api.APIError(400, 'invalid-field', field='timestamps')

    if timestamps is not None:
        samples = await core_history.get_samples_by_timestamp(port, timestamps)

    else:
        samples = await core_history.get_samples_slice(port, from_timestamp, to_timestamp, limit)

    return list(samples)


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
async def delete_port_history(request: core_api.APIRequest, port_id: str) -> None:
    port = core_ports.get(port_id)
    if port is None:
        raise core_api.APIError(404, 'no-such-port')

    query = request.query

    from_str = query.get('from')
    if from_str is None:
        raise core_api.APIError(400, 'missing-field', field='from')

    try:
        from_timestamp = int(from_str)

    except ValueError:
        raise core_api.APIError(400, 'invalid-field', field='from') from None

    if from_timestamp < 0:
        raise core_api.APIError(400, 'invalid-field', field='from')

    to_str = query.get('to')
    if to_str is None:
        raise core_api.APIError(400, 'missing-field', field='to')

    try:
        to_timestamp = int(to_str)

    except ValueError:
        raise core_api.APIError(400, 'invalid-field', field='to') from None

    if to_timestamp < 0:
        raise core_api.APIError(400, 'invalid-field', field='to')

    await core_history.remove_samples([port], from_timestamp, to_timestamp)




from __future__ import annotations

import abc

from typing import Optional, Set, Union


class Expression(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    async def eval(self) -> Evaluated:
        raise NotImplementedError()

    def get_deps(self) -> Set[str]:
        # Special deps:
        #  * 'second' - used to indicate dependency on system time (seconds)
        #  * 'millisecond' - used to indicate dependency on system time (milliseconds)

        return set()

    @staticmethod
    @abc.abstractmethod
    def parse(self_port_id: Optional[str], sexpression: str, pos: int) -> Expression:
        raise NotImplementedError()


# This needs to be imported here to avoid circular import issues
from qtoggleserver.core import ports as core_ports  # noqa: E402

Evaluated = Union[bool, int, float, core_ports.BasePort]




import logging

from typing import Dict, Tuple

from qtoggleserver.core import events as core_events
from qtoggleserver.core import ports as core_ports
from qtoggleserver.core.typing import Attribute, Attributes, NullablePortValue
from qtoggleserver.lib import filtereventhandler
from qtoggleserver.slaves import devices as slaves_devices


logger = logging.getLogger(__name__)


class DummyEventHandler(filtereventhandler.FilterEventHandler):
    async def on_event(self, event: core_events.Event) -> None:
        logger.debug('handling %s', event)

    async def on_value_change(
        self,
        event: core_events.Event,
        port: core_ports.BasePort,
        old_value: NullablePortValue,
        new_value: NullablePortValue,
        attrs: Attributes
    ) -> None:

        logger.debug('handling value-change for %s', port)

    async def on_port_update(
        self,
        event: core_events.Event,
        port: core_ports.BasePort,
        old_attrs: Attributes,
        new_attrs: Attributes,
        changed_attrs: Dict[str, Tuple[Attribute, Attribute]],
        added_attrs: Attributes,
        removed_attrs: Attributes
    ) -> None:

        logger.debug('handling port-update for %s', port)

    async def on_port_add(self, event: core_events.Event, port: core_ports.BasePort, attrs: Attributes) -> None:
        logger.debug('handling port-add for %s', port)

    async def on_port_remove(self, event: core_events.Event, port: core_ports.BasePort, attrs: Attributes) -> None:
        logger.debug('handling port-remove for %s', port)

    async def on_device_update(
        self,
        event: core_events.Event,
        old_attrs: Attributes,
        new_attrs: Attributes,
        changed_attrs: Dict[str, Tuple[Attribute, Attribute]],
        added_attrs: Attributes,
        removed_attrs: Attributes
    ) -> None:

        logger.debug('handling device-update')

    async def on_full_update(self, event: core_events.Event) -> None:
        logger.debug('handling full-update')

    async def on_slave_device_update(
        self,
        event: core_events.Event,
        slave: slaves_devices.Slave,
        old_attrs: Attributes,
        new_attrs: Attributes,
        changed_attrs: Dict[str, Tuple[Attribute, Attribute]],
        added_attrs: Attributes,
        removed_attrs: Attributes
    ) -> None:

        logger.debug('handling slave-device-update for %s', slave)

    async def on_slave_device_add(
        self,
        event: core_events.Event,
        slave: slaves_devices.Slave,
        attrs: Attributes
    ) -> None:

        logger.debug('handling slave-device-add for %s', slave)

    async def on_slave_device_remove(
        self,
        event: core_events.Event,
        slave: slaves_devices.Slave, attrs: Attributes
    ) -> None:

        logger.debug('handling slave-device-remove for %s', slave)




import asyncio
import datetime
import logging
import re

from typing import Any, AsyncContextManager, Dict, Iterable, List, Optional, Set, Tuple

import asyncpg.pool

from qtoggleserver.persist import BaseDriver
from qtoggleserver.persist.typing import Id, Record
from qtoggleserver.utils import json as json_utils


logger = logging.getLogger(__name__)

DEFAULT_DB = 'qtoggleserver'
POOL_MIN_CONNECTIONS = 2
POOL_MAX_CONNECTIONS = 4
POOL_MAX_QUERIES = 256

FILTER_OP_MAPPING = {
    'gt': '>',
    'ge': '>=',
    'lt': '<',
    'le': '<=',
    'in': 'in'
}

D_FMT = '__{:04d}-{:02d}-{:02d}T'
D_FMT_LEN = 13
D_REGEX = re.compile(r'^__(\d{4})-(\d{2})-(\d{2})T$')

DT_FMT = '__{:04d}-{:02d}-{:02d}T{:02d}:{:02d}:{:02d}.{:06d}Z'
DT_FMT_LEN = 29
DT_REGEX = re.compile(r'^__(\d{4})-(\d{2})-(\d{2})T(\d{2}):(\d{2}):(\d{2}).(\d{6})Z$')


class PostgreSQLDriver(BaseDriver):
    def __init__(
        self,
        host: str = '127.0.0.1',
        port: str = 5432,
        db: str = DEFAULT_DB,
        username: Optional[str] = None,
        password: Optional[str] = None,
        **kwargs
    ) -> None:
        logger.debug('using %s:%s/%s', host, port, db)

        conn_details = {
            'host': host,
            'port': port,
            'dbname': db
        }

        if username:
            conn_details['user'] = username
        if password:
            conn_details['password'] = password

        self._conn_details = {
            'user': username,
            'password': password,
            'database': db,
            'host': host,
            'port': port
        }

        self._conn_pool: Optional[asyncpg.pool.Pool] = None
        self._existing_tables: Optional[Set[str]] = None
        self._ensure_table_exists_lock: asyncio.Lock = asyncio.Lock()

    async def query(
        self,
        collection: str,
        fields: Optional[List[str]],
        filt: Dict[str, Any],
        sort: List[Tuple[str, bool]],
        limit: Optional[int]
    ) -> Iterable[Record]:

        await self._ensure_table_exists(collection)

        params = []
        if fields:
            select_clause = self._fields_to_select_clause(fields, params)

        else:
            select_clause = 'id, content'

        query = f'SELECT {select_clause} FROM {collection}'

        where_clause = self._filt_to_where_clause(self._filt_to_db(filt), params)
        if where_clause:
            query += f' WHERE {where_clause}'

        if sort:
            order_by_clause = self._fields_to_order_by_clause(sort, params)
            query += f' ORDER BY {order_by_clause}'

        if limit is not None:
            query += f' LIMIT ${len(params) + 1}'
            params.append(limit)

        results = await self._execute_query(query, params)

        return self._query_gen_wrapper(results, fields)

    async def insert(self, collection: str, record: Record) -> Id:
        await self._ensure_table_exists(collection)

        db_record = self._record_to_db(record)
        id_ = db_record.pop('id', None)

        if id_ is not None:
            statement = f'INSERT INTO {collection}(id, content) VALUES($1, $2) RETURNING id'
            params = [id_, db_record]

        else:
            statement = f'INSERT INTO {collection}(content) VALUES($1) RETURNING id'
            params = [db_record]

        _, result_rows = await self._execute_statement(statement, params, has_result_rows=True)

        return result_rows[0][0]

    async def update(self, collection: str, record_part: Record, filt: Dict[str, Any]) -> int:
        await self._ensure_table_exists(collection)

        params = []
        statement = f'UPDATE {collection}'

        db_record_part = self._record_to_db(record_part)
        update_clause = self._record_to_update_clause(db_record_part, params)
        statement += f' SET {update_clause}'

        where_clause = self._filt_to_where_clause(self._filt_to_db(filt), params)
        if where_clause:
            statement += f' WHERE {where_clause}'

        status_msg, _ = await self._execute_statement(statement, params)

        count = int(status_msg.split()[1])  # Assuming status_msg has format "UPDATE ${count}"

        return count

    async def replace(self, collection: str, id_: Id, record: Record) -> bool:
        await self._ensure_table_exists(collection)

        db_record = self._record_to_db(record)
        statement = f'UPDATE {collection} SET content = $1 WHERE id = $2'
        params = [db_record, id_]

        status_msg, _ = await self._execute_statement(statement, params)

        count = int(status_msg.split()[1])  # Assuming status_msg has format "UPDATE ${count}"

        return count > 0

    async def remove(self, collection: str, filt: Dict[str, Any]) -> int:
        await self._ensure_table_exists(collection)

        params = []
        statement = f'DELETE FROM {collection}'

        where_clause = self._filt_to_where_clause(self._filt_to_db(filt), params)
        if where_clause:
            statement += f' WHERE {where_clause}'

        status_msg, _ = await self._execute_statement(statement, params)

        count = int(status_msg.split()[1])  # Assuming status_msg has format "DELETE ${count}"

        return count

    async def ensure_index(self, collection: str, index: List[Tuple[str, bool]]) -> None:
        await self._ensure_table_exists(collection)

        field_names = [i[0] for i in index]
        field_names_str = '_'.join(field_names)
        index_name = f'{collection}_{field_names_str}'

        params = []
        index_statement = self._index_to_index_clause(index, params)
        statement = f'CREATE INDEX IF NOT EXISTS {index_name} ON {collection}({index_statement})'

        await self._execute_statement(statement, params)

    async def cleanup(self) -> None:
        logger.debug('disconnecting client')

        if self._conn_pool:
            await self._conn_pool.close()
            self._conn_pool = None

    def is_history_supported(self) -> bool:
        return True

    async def _acquire_connection(self) -> AsyncContextManager[asyncpg.Connection]:
        if self._conn_pool is None:
            logger.debug('creating connection pool')
            self._conn_pool = await asyncpg.create_pool(
                min_size=POOL_MIN_CONNECTIONS,
                max_size=POOL_MAX_CONNECTIONS,
                max_queries=POOL_MAX_QUERIES,
                init=self._init_connection,
                **self._conn_details
            )

        return self._conn_pool.acquire()

    async def _init_connection(self, conn: asyncpg.Connection) -> None:
        # Install JSON converter
        await conn.set_type_codec(
            'jsonb',
            encoder=json_utils.dumps,
            decoder=json_utils.loads,
            schema='pg_catalog',
            format='text'
        )

    async def _ensure_table_exists(self, table_name: str) -> None:
        async with self._ensure_table_exists_lock:
            if self._existing_tables is None:
                self._existing_tables = await self._get_existing_table_names()

            if table_name in self._existing_tables:
                return

            statement = f'CREATE SEQUENCE {table_name}_id_seq'
            await self._execute_statement(statement)

            id_data_type = f"TEXT NOT NULL DEFAULT NEXTVAL('{table_name}_id_seq') PRIMARY KEY"
            statement = f'CREATE TABLE {table_name}(id {id_data_type}, content JSONB)'
            await self._execute_statement(statement)

            self._existing_tables.add(table_name)

    async def _get_existing_table_names(self) -> Set[str]:
        query = (
            "SELECT table_name "
            "FROM information_schema.tables "
            "WHERE table_type = 'BASE TABLE' AND table_schema = 'public'"
        )

        results = await self._execute_query(query)

        return set(r[0] for r in results)

    async def _execute_query(self, query: str, params: Iterable[Any] = None) -> Iterable[tuple]:
        async with await self._acquire_connection() as conn:
            async with conn.transaction():
                return [row async for row in conn.cursor(query, *(params or []))]

    async def _execute_statement(
        self,
        statement: str,
        params: Iterable[Any] = None,
        has_result_rows: bool = False
    ) -> Tuple[str, List[tuple]]:
        async with await self._acquire_connection() as conn:
            stmt = await conn.prepare(statement)

            async with conn.transaction():
                rows = await stmt.fetch(*(params or []))
                if has_result_rows:
                    return stmt.get_statusmsg(), rows

                else:
                    return stmt.get_statusmsg(), []

    @staticmethod
    def _fields_to_select_clause(fields: List[str], params: List[Any]) -> str:
        select_clause = []

        for field in fields:
            if field == 'id':
                select_clause.append('id')

            else:
                select_clause.append(f'content->${len(params) + 1}')
                params.append(field)

        select_clause = ', '.join(select_clause)

        return select_clause

    @staticmethod
    def _fields_to_order_by_clause(fields: List[Tuple[str, bool]], params: List[Any]) -> str:
        order_by_clause = []

        for field, desc in fields:
            if field == 'id':
                clause = "SUBSTRING(id FROM '([0-9]+)')::int"

            else:
                clause = f'content->${len(params) + 1}'
                params.append(field)

            clause = f"{clause} {'DESC' if desc else 'ASC'}"
            order_by_clause.append(clause)

        order_by_clause = ', '.join(order_by_clause)

        return order_by_clause

    @staticmethod
    def _filt_to_where_clause(filt: Dict[str, Any], params: List[Any]) -> str:
        where_clause = []

        for key, value in filt.items():
            if isinstance(value, dict):  # filter with operators
                ops_values = [(FILTER_OP_MAPPING[k], v) for k, v in value.items()]

            else:
                ops_values = [('=', value)]

            if key == 'id':
                for o, v in ops_values:
                    if o == 'in':
                        placeholder = ', '.join(f'${len(params) + 1 + i}' for i, _ in enumerate(v))
                        placeholder = f'({placeholder})'
                        params += v

                    else:
                        placeholder = f'${len(params) + 1}'
                        params.append(v)

                    condition = f'{key} {o} {placeholder}'
                    where_clause.append(condition)

            else:
                for o, v in ops_values:
                    if o == 'in':
                        placeholder = ', '.join(f'${len(params) + 1 + i}' for i, _ in enumerate(v))
                        placeholder = f'({placeholder})'
                        params += v

                    else:
                        placeholder = f'${len(params) + 1}'
                        params.append(v)

                    condition = f"content->'{key}' {o} {placeholder}"
                    where_clause.append(condition)

        where_clause = ' AND '.join(where_clause)

        return where_clause

    @staticmethod
    def _record_to_update_clause(record: Record, params: List[Any]) -> str:
        update_clause = []

        id_ = record.pop('id', None)
        if id_ is not None:
            update_clause.append(f'id = ${len(params) + 1}')
            params.append(id_)

        if record:
            update_clause.append(f'content = content || ${len(params) + 1}')
            params.append(record)

        update_clause = ', '.join(update_clause)

        return update_clause

    @staticmethod
    def _index_to_index_clause(index: List[Tuple[str, bool]], params: List[Any]) -> str:
        index_clause = []

        # We can't use prepared statements with CREATE INDEX, so we need to build our query without arguments
        for field, _ in index:
            field = re.sub(r'[^a-zA-Z0-9_]', '', field)
            index_clause.append(f"(content->'${field}')")

        index_clause = ', '.join(index_clause)

        return index_clause

    def _query_gen_wrapper(self, results: Iterable[Tuple[Any]], fields: Optional[List[str]]) -> Iterable[Record]:
        if fields:
            for result in results:
                db_record = {fields[i]: r for i, r in enumerate(result)}
                yield self._record_from_db(db_record)

        else:
            for result in results:
                id_, db_record = result
                db_record['id'] = id_
                yield self._record_from_db(db_record)

    def _filt_to_db(self, filt: Dict[str, Any]) -> Dict[str, Any]:
        db_filt = {}
        for key, value in filt.items():
            if isinstance(value, dict):  # filter with operators
                value = {k: self.value_to_db(v) for k, v in value.items()}

            else:
                value = self.value_to_db(value)

            db_filt[key] = value

        return db_filt

    def _record_to_db(self, record: Record) -> Record:
        return {k: self.value_to_db(v) for k, v in record.items()}

    def _record_from_db(self, record: Record) -> Record:
        return {k: self.value_from_db(v) for k, v in record.items()}

    def value_to_db(self, value: Any) -> Any:
        if isinstance(value, datetime.datetime):
            return self.datetime_to_str(value)

        if isinstance(value, datetime.date):
            return self.date_to_str(value)

        return value

    def value_from_db(self, value: Any) -> Any:
        if isinstance(value, str):
            return self.str_to_datetime(value) or self.str_to_date(value) or value

        return value

    @staticmethod
    def datetime_to_str(dt: datetime.datetime) -> str:
        return DT_FMT.format(dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second, dt.microsecond)

    @staticmethod
    def str_to_datetime(s: str) -> Optional[datetime.datetime]:
        if len(s) != DT_FMT_LEN:
            return

        m = DT_REGEX.match(s)
        if m is None:
            return

        return datetime.datetime(*(int(g) for g in m.groups()))

    @staticmethod
    def date_to_str(dt: datetime.date) -> str:
        return D_FMT.format(dt.year, dt.month, dt.day)

    @staticmethod
    def str_to_date(s: str) -> Optional[datetime.date]:
        if len(s) != D_FMT_LEN:
            return

        m = D_REGEX.match(s)
        if m is None:
            return

        return datetime.date(*(int(g) for g in m.groups()))




import time

from typing import Any, Dict, Optional

from qtoggleserver import system
from qtoggleserver.core import history

from .base import Evaluated
from .exceptions import PortValueUnavailable, ExpressionEvalError, EvalSkipped
from .functions import function, Function
from .port import PortRef


@function('AVAILABLE')
class AvailableFunction(Function):
    MIN_ARGS = MAX_ARGS = 1

    async def eval(self) -> Evaluated:
        try:
            await self.args[0].eval()
            return True

        except ExpressionEvalError:
            return False


@function('DEFAULT')
class DefaultFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    async def eval(self) -> Evaluated:
        try:
            return await self.args[0].eval()

        except ExpressionEvalError:
            return await self.args[1].eval()


@function('ACC')
class AccFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._last_value: Optional[float] = None

    async def eval(self) -> Evaluated:
        value, accumulator = await self.eval_args()
        result = accumulator

        if self._last_value is not None:
            result += value - self._last_value

        self._last_value = value

        return result


@function('ACCINC')
class AccIncFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._last_value: Optional[float] = None

    async def eval(self) -> Evaluated:
        value, accumulator = await self.eval_args()
        result = accumulator

        if (self._last_value is not None) and (value > self._last_value):
            result += value - self._last_value

        self._last_value = value

        return result


@function('HYST')
class HystFunction(Function):
    MIN_ARGS = MAX_ARGS = 3

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._last_result: int = 0

    async def eval(self) -> Evaluated:
        value, threshold1, threshold2 = await self.eval_args()

        self._last_result = int((self._last_result == 0 and value > threshold2) or
                                (self._last_result != 0 and value >= threshold1))

        return self._last_result


@function('SEQUENCE')
class SequenceFunction(Function):
    MIN_ARGS = 2
    DEPS = ['millisecond']

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._last_time: float = 0

    async def eval(self) -> Evaluated:
        now = time.time() * 1000

        if self._last_time == 0:
            self._last_time = now

        args = await self.eval_args()
        num_values = len(args) // 2
        values = []
        delays = []
        total_delay = 0
        for i in range(num_values):
            values.append(args[i * 2])
            delays.append(args[i * 2 + 1])
            total_delay += args[i * 2 + 1]

        if len(delays) < len(values):
            delays.append(0)

        delta = now - self._last_time
        delta = delta % total_delay  # Work modulo total_delay, to create repeat effect
        delay_so_far = 0
        result = values[0]
        for i in range(num_values):
            delay_so_far += delays[i]
            if delay_so_far >= delta:
                result = values[i]
                break

        return result


@function('LUT')
class LUTFunction(Function):
    MIN_ARGS = 5

    async def eval(self) -> Evaluated:
        args = await self.eval_args()
        length = (len(args) - 1) // 2
        x = args[0]
        points = [(args[2 * i + 1], args[2 * i + 2]) for i in range(length)]
        points.sort(key=lambda p: p[0])

        if x < points[0][0]:
            return points[0][1]

        for i in range(length - 1):
            p1 = points[i]
            p2 = points[i + 1]

            if x > p2[0]:
                continue

            if x - p1[0] < p2[0] - x:  # Closer to p1 than to p2
                return p1[1]

            else:
                return p2[1]

        return points[length - 1][1]


@function('LUTLI')
class LUTLIFunction(Function):
    MIN_ARGS = 5

    async def eval(self) -> Evaluated:
        args = await self.eval_args()
        length = (len(args) - 1) // 2
        x = args[0]
        points = [(args[2 * i + 1], args[2 * i + 2]) for i in range(length)]
        points.sort(key=lambda p: p[0])

        if x < points[0][0]:
            return points[0][1]

        for i in range(length - 1):
            p1 = points[i]
            p2 = points[i + 1]

            if x > p2[0]:
                continue

            if p1[0] == p2[0]:
                return p1[1]

            return p1[1] + (p2[1] - p1[1]) * (x - p1[0]) / (p2[0] - p1[0])

        return points[length - 1][1]


@function('HISTORY')
class HistoryFunction(Function):
    MIN_ARGS = MAX_ARGS = 3
    DEPS = ['second']
    ARG_KINDS = [PortRef]
    ENABLED = history.is_enabled

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._cached_sample: Optional[Dict[str, Any]] = None
        self._cached_timestamp: int = 0
        self._cached_max_diff: Optional[float] = None

    async def eval(self) -> Evaluated:
        if not system.date.has_real_date_time():
            raise EvalSkipped()

        args = await self.eval_args()
        port, timestamp, max_diff = args

        # Transform everything to milliseconds
        timestamp *= 1000
        max_diff *= 1000
        now_ms = int(time.time() * 1000)

        # If arguments have changed from last cached values, invalidate the sample
        if self._cached_timestamp != timestamp or self._cached_max_diff != max_diff:
            self._cached_sample = None

        if self._cached_sample is not None:
            return self._cached_sample['value']

        if max_diff > 0:
            # Look through all values after given timestamp, but no newer than timestamp + max_diff, and consider the
            # oldest one
            from_timestamp = timestamp
            to_timestamp = timestamp + max_diff
            sort_desc = False
            consider_curr_value = from_timestamp <= now_ms < to_timestamp

        elif max_diff < 0:
            # Look through all values before given timestamp, but no older than timestamp - abs(max_diff), and consider
            # the newest one
            if now_ms <= timestamp:
                # Sample from the future requested, the best we've got is current value
                from_timestamp = to_timestamp = None
                consider_curr_value = True

            else:
                # +1 is needed to satisfy the inclusion/exclusion of the interval boundaries
                from_timestamp = timestamp + max_diff + 1  # max_diff is negative
                to_timestamp = timestamp + 1
                consider_curr_value = False

            sort_desc = True

        else:  # Assuming max_dif == 0
            # Look through all values after given timestamp and consider the oldest one
            from_timestamp = timestamp
            to_timestamp = None
            sort_desc = False
            consider_curr_value = from_timestamp <= now_ms

        if from_timestamp is not None or to_timestamp is not None:
            samples = await history.get_samples_slice(port, from_timestamp, to_timestamp, limit=1, sort_desc=sort_desc)
            samples = list(samples)

        else:
            samples = []

        if samples:
            self._cached_sample = samples[0]
            self._cached_timestamp = timestamp
            self._cached_max_diff = max_diff
            value = self._cached_sample['value']

        elif consider_curr_value:
            value = port.get_last_read_value()

        else:
            value = None

        if value is None:
            raise PortValueUnavailable(port.get_id())

        return value




import logging
import subprocess

from typing import Dict, List, Optional


logger = logging.getLogger(__name__)


def run_get_cmd(
    get_cmd: str,
    cmd_name: Optional[str] = None,
    log_values: bool = True,
    exc_class: type = None,
    required_fields: Optional[List[str]] = None
) -> Dict[str, str]:

    exc_class = exc_class or Exception

    try:
        config = subprocess.check_output(get_cmd, stderr=subprocess.STDOUT, shell=True)

    except Exception as e:
        raise exc_class(f'{cmd_name or get_cmd} get command failed: {e}') from e

    config = config.strip().decode()
    config_lines = config.split('\n')
    config_dict = {}
    for line in config_lines:
        line = line.strip()
        if not line:
            continue

        parts = line.split('=', 1)
        if len(parts) == 1:
            parts.append('')

        key, value = parts
        key = key.lower()[3:]  # Strip leading "QS_"
        if value.startswith('"'):
            value = value[1:]
        if value.endswith('"') and not value.endswith('\\"'):
            value = value[:-1]

        config_dict[key] = value

    if cmd_name:
        if log_values:
            values_str = ', '.join(f'{k} = "{v}"' for k, v in sorted(config_dict.items()))
            logger.debug('got %s: %s', cmd_name, values_str)

        else:
            logger.debug('got %s', cmd_name)

    for field in required_fields or []:
        if field not in config_dict:
            msg = f'missing {field} field'
            if cmd_name:
                msg = f'invalid {cmd_name}: {msg}'

            raise exc_class(msg)

    return config_dict


def run_set_cmd(
    set_cmd: str,
    cmd_name: Optional[str] = None,
    log_values: bool = True,
    exc_class: Optional[type] = None,
    **config
) -> None:

    env = {f'QS_{k.upper()}': v for k, v in config.items()}

    exc_class = exc_class or Exception

    try:
        subprocess.check_output(set_cmd, env=env, stderr=subprocess.STDOUT, shell=True)

    except Exception as e:
        raise exc_class(f'{cmd_name or set_cmd} set command failed: {e}') from e

    if cmd_name:
        if log_values:
            values_str = ', '.join(f'{k} = "{v}"' for k, v in sorted(config.items()))
            logger.debug('%s set to: %s', cmd_name, values_str)

        else:
            logger.debug('%s set', cmd_name)




import os.path

from setuptools import setup, find_namespace_packages

import qtoggleserver.version

try:
    import fastentrypoints

except ImportError:
    pass

try:
    import setupnovernormalize

except ImportError:
    pass


name = 'qtoggleserver'
version = qtoggleserver.version.VERSION
here = os.path.dirname(__file__) or '.'


setup(
    name=name,
    version=version,
    description='A fully fledged qToggle Python implementation',
    author='The qToggle Team',
    author_email='ccrisan@gmail.com',
    url='https://github.com/qtoggle/qtoggleserver',
    license='Apache 2.0',

    packages=find_namespace_packages(
        include=[name, f'{name}.*'],
        exclude=[f'{name}.frontend.node_modules.*']
    ),
    namespace_packages=[name],
    py_modules=[f'{name}.{module}' for module in [
        'persist',
        'version'
    ]],

    install_requires=[
        'jinja2>=2.10,<3',
        'jsonpointer>=2.0,<3',
        'jsonschema>=2.5,<4',
        'psutil>=5.5,<6',
        'pyhocon>=0.3.50,<1',
        'pyjwt>=1.7,<2',
        'pytz',
        'qui-server>=1.13.3',
        'tornado>=6.0,<7',
    ],

    zip_safe=False,
    include_package_data=False,
    package_data={
        name: [
            'frontend/dist/font/*',
            'frontend/dist/templates/*',
            'frontend/dist/img/*',
            'frontend/dist/*'
        ]
    },
    data_files=[
        (f'share/{name}/{root[len(here) + 1:]}', [f'{root[len(here) + 1:]}/{f}' for f in files])
        for (root, dirs, files) in os.walk(os.path.join(here, 'extra'))
    ],

    entry_points={
        'console_scripts': [
            'qtoggleserver=qtoggleserver.commands.server:execute',
            'qtoggleshell=qtoggleserver.commands.shell:execute'
        ],
    }
)




import asyncio
import inspect
import logging
import re

from typing import Any, Callable, Dict, Optional

from tornado.iostream import StreamClosedError
from tornado.web import RequestHandler, HTTPError


from qtoggleserver.core import api as core_api
from qtoggleserver.core import responses as core_responses
from qtoggleserver.core.device import attrs as core_device_attrs
from qtoggleserver.core.api import auth as core_api_auth
from qtoggleserver.utils import json as json_utils


SESSION_ID_RE = re.compile(r'[a-zA-Z0-9]{1,32}')

logger = logging.getLogger(__name__)


class NoSuchFunction(HTTPError):
    def __init__(self) -> None:
        super().__init__(404, 'no-such-function')


class BaseHandler(RequestHandler):
    _UNDEFINED = {}

    def __init__(self, *args, **kwargs) -> None:
        self._json: Any = self._UNDEFINED
        self._response_body: str = ''
        self._response_body_json: Any = None

        super().__init__(*args, **kwargs)

    def get_request_json(self) -> Any:
        if self._json is self._UNDEFINED:
            try:
                self._json = json_utils.loads(self.request.body)

            except ValueError as e:
                logger.error('could not decode json from request body: %s', e)

                raise core_api.APIError(400, 'malformed-body') from e

        return self._json

    def finish(self, chunk: Optional[str] = None) -> asyncio.Future:
        self._response_body = chunk

        return super().finish(chunk)

    def finish_json(self, data: Any) -> asyncio.Future:
        self._response_body_json = data

        data = json_utils.dumps(data)
        data += '\n'

        self.set_header('Content-Type', 'application/json; charset=utf-8')
        return self.finish(data)

    def get_response_body(self) -> str:
        return self._response_body

    def get_response_body_json(self) -> Any:
        return self._response_body_json

    def get_response_headers(self) -> Dict[str, str]:
        return dict(self._headers.get_all())

    def get(self, **kwargs) -> None:
        raise NoSuchFunction()

    head = post = delete = patch = put = options = get

    def _handle_request_exception(self, exception: Exception) -> None:
        try:
            if isinstance(exception, HTTPError):
                logger.error('%s %s: %s', self.request.method, self.request.uri, exception)
                self.set_status(exception.status_code)
                self.finish_json({'error': (exception.log_message or
                                            getattr(exception, 'reason', None) or str(exception))})

            else:
                logger.error(str(exception), exc_info=True)
                self.set_status(500)
                self.finish_json({'error': 'internal server error'})

        except RuntimeError:
            pass  # Nevermind

    def data_received(self, chunk: bytes) -> None:
        pass


class APIHandler(BaseHandler):
    AUTH_ENABLED = True

    def __init__(self, *args, **kwargs) -> None:
        self.access_level: int = core_api.ACCESS_LEVEL_NONE
        self.username: Optional[str] = None

        super().__init__(*args, **kwargs)

    def prepare(self) -> None:
        # Disable cache
        self.set_header('Cache-Control', 'no-cache, no-store, must-revalidate, max-age=0')

        if not self.AUTH_ENABLED:
            return

        # Parse auth header
        auth = self.request.headers.get('Authorization')
        if auth:
            try:
                usr = core_api_auth.parse_auth_header(
                    auth,
                    core_api_auth.ORIGIN_CONSUMER,
                    core_api_auth.consumer_password_hash_func
                )

            except core_api_auth.AuthError as e:
                logger.warning(str(e))
                return

        else:
            if core_device_attrs.admin_password_hash == core_device_attrs.EMPTY_PASSWORD_HASH:
                logger.debug('authenticating request as admin due to empty admin password')
                usr = 'admin'

            else:
                logger.warning('missing authorization header')
                return

        self.access_level = core_api.ACCESS_LEVEL_MAPPING[usr]
        self.username = usr

        logger.debug(
            'granted access level %s (username=%s)',
            core_api.ACCESS_LEVEL_MAPPING[self.access_level],
            self.username
        )

        # Validate session id
        session_id = self.request.headers.get('Session-Id')
        if session_id:
            if not SESSION_ID_RE.match(session_id):
                raise core_api.APIError(400, 'invalid-header', header='Session-Id')

    async def call_api_func(self, func: Callable, default_status: int = 200, **kwargs) -> None:
        try:
            if (self.request.method in ('POST', 'PATCH', 'PUT') and
                self.request.headers.get('Content-Type', '').startswith('application/json')):

                kwargs['params'] = self.get_request_json()

            response = func(self, **kwargs)
            if inspect.isawaitable(response):
                response = await response

            self.set_status(default_status)
            if response is not None or default_status == 200:
                await self.finish_json(response)

            else:
                await self.finish()

        except Exception as e:
            await self._handle_api_call_exception(func, kwargs, e)

    async def _handle_api_call_exception(self, func: Callable, kwargs: dict, error: Exception) -> None:
        kwargs = dict(kwargs)
        params = kwargs.pop('params', None)
        args = json_utils.dumps(kwargs)
        body = params and json_utils.dumps(params) or '{}'

        if isinstance(error, core_responses.HTTPError):
            error = core_api.APIError.from_http_error(error)

        if isinstance(error, core_api.APIError):
            logger.error('api call %s failed: %s (args=%s, body=%s)', func.__name__, error, args, body)

            self.set_status(error.status)
            if not self._finished:  # Avoid finishing an already finished request
                await self.finish_json(error.to_json())

        if isinstance(error, core_api.APIAccepted):
            self.set_status(202)
            if not self._finished and error.response is not None:  # Avoid finishing an already finished request
                await self.finish_json(error.response)

        elif isinstance(error, StreamClosedError) and func.__name__ == 'get_listen':
            logger.debug('api call get_listen could not complete: stream closed')

        else:
            logger.error('api call %s failed: %s (args=%s, body=%s)', func.__name__, error, args, body, exc_info=True)

            self.set_status(500)
            if not self._finished:  # Avoid finishing an already finished request
                await self.finish_json({'error': str(error)})




from typing import Any, Dict

import logging
import os

from qtoggleserver.conf import settings
from qtoggleserver.utils import conf as conf_utils


logger = logging.getLogger(__name__)


def can_write_conf_file() -> bool:
    if not settings.source:
        return False

    return os.access(settings.source, os.W_OK)


def conf_file_to_dict() -> Dict[str, Any]:
    if not can_write_conf_file():
        raise Exception('Configuration file not available')

    return conf_utils.config_from_file(settings.source)


def conf_file_from_dict(d: Dict[str, Any]) -> None:
    if not can_write_conf_file():
        raise Exception('Configuration file not available')

    logger.debug('updating configuration file %s', settings.source)

    config = conf_utils.config_from_dict(d)
    config_str = conf_utils.config_to_str(config)

    with open(settings.source, 'wt') as f:
        f.write(config_str)




import asyncio
import logging
import time

from typing import Dict, Optional, Set, Union

from qtoggleserver.conf import settings
from qtoggleserver.core import expressions as core_expressions
from qtoggleserver.core import ports as core_ports
from qtoggleserver.utils import json as json_utils
from qtoggleserver.utils import timedset
from qtoggleserver.utils import logging as logging_utils


# After how much time to retry reading a port whose read_value() method raised an error
_PORT_READ_ERROR_RETRY_INTERVAL = 10

logger = logging.getLogger(__name__)
memory_logs: Optional[logging_utils.FifoMemoryHandler] = None

loop: Optional[asyncio.AbstractEventLoop] = None

_update_loop_task: Optional[asyncio.Task] = None
_ready: bool = False
_updating_enabled: bool = True
_start_time: float = time.time()
_last_time: float = 0
_force_eval_expression_ports: Set[Union[core_ports.BasePort, None]] = set()
_ports_with_read_error = timedset.TimedSet(_PORT_READ_ERROR_RETRY_INTERVAL)
_update_lock: Optional[asyncio.Lock] = None


async def update() -> None:
    from . import ports
    from . import sessions

    global _last_time
    global _update_lock

    if not _updating_enabled:
        return

    if _update_lock is None:
        _update_lock = asyncio.Lock()

    async with _update_lock:
        changed_set = {'millisecond'}
        change_reasons = {}

        now = int(time.time())
        time_changed = False
        if now != _last_time:
            _last_time = now
            time_changed = True
            changed_set.add('second')

        for port in ports.get_all():
            if not port.is_enabled():
                continue

            port.invalidate_attrs()
            old_value = port.get_last_read_value()

            try:
                port.heart_beat()

            except Exception as e:
                logger.error('port heart beat exception: %s', e, exc_info=True)

            if time_changed:
                try:
                    port.heart_beat_second()

                except Exception as e:
                    logger.error('port heart beat second exception: %s', e, exc_info=True)

            # Skip ports with read errors for a while
            if port in _ports_with_read_error:
                continue

            try:
                new_value = await port.read_transformed_value()

            except Exception as e:
                logger.error('failed to read value from %s: %s', port, e, exc_info=True)
                _ports_with_read_error.add(port)

                continue

            if new_value is None:
                continue  # Read value not available

            if new_value != old_value:
                old_value_str = json_utils.dumps(old_value) if old_value is not None else '(unavailable)'
                new_value_str = json_utils.dumps(new_value)

                logger.debug('detected %s value change: %s -> %s', port, old_value_str, new_value_str)

                port.set_last_read_value(new_value)
                changed_set.add(port)

                # Remember and reset port change reason
                change_reasons[port] = port.get_change_reason()
                port.reset_change_reason()

        await handle_value_changes(changed_set, change_reasons)

        sessions.update()


async def update_loop() -> None:
    while True:
        try:
            try:
                if _ready:
                    await update()

            except Exception as e:
                logger.error('update failed: %s', e, exc_info=True)

            await asyncio.sleep(settings.core.tick_interval / 1000.0)

        except asyncio.CancelledError:
            logger.debug('update loop cancelled')
            break


async def handle_value_changes(
    changed_set: Set[Union[core_ports.BasePort, None]],
    change_reasons: Dict[core_ports.BasePort, str]
) -> None:

    if _force_eval_expression_ports:
        changed_set.update(_force_eval_expression_ports)  # Special "always depends on" value
        _force_eval_expression_ports.clear()

    # Trigger value-change events; save persisted ports
    for port in changed_set:
        if not isinstance(port, core_ports.BasePort):
            continue

        if not await port.is_internal():
            await port.trigger_value_change()

        if await port.is_persisted():
            await port.save()

    # Reevaluate the expressions depending on changed ports
    for port in core_ports.get_all():
        if not port.is_enabled():
            continue

        expression = await port.get_expression()
        if not expression:
            continue

        deps = set(expression.get_deps())
        deps.add(None)  # Special "always depends on" value

        changed_set_ids = set(f'${c.get_id()}' for c in changed_set if isinstance(c, core_ports.BasePort))

        # Evaluate a port's expression when:
        # * one of its deps changed
        # * its own value changed

        # Join all deps together; deps may contain:
        # * ports
        # * port ids
        # * time dep strings
        # * None
        if not (deps & (changed_set_ids | changed_set)) and (port not in changed_set):
            continue

        # If port expression depends on port itself and the change reason is the evaluation of its expression, prevent
        # evaluating its expression again to avoid evaluation loops
        change_reason = change_reasons.get(port, core_ports.CHANGE_REASON_NATIVE)
        if ((port in changed_set) and
            (f'${port.get_id()}' in deps) and
            (change_reason == core_ports.CHANGE_REASON_EXPRESSION)):

            logger.debug('skipping evaluation of %s expression to prevent loops', port)
            continue

        try:
            value = await expression.eval()

        except core_expressions.ExpressionEvalError:
            continue

        except Exception as e:
            logger.error('failed to evaluate expression "%s" of %s: %s', expression, port, e)
            continue

        value = await port.adapt_value_type(value)
        if value is None:
            continue

        if value != port.get_last_read_value():  # Value changed after evaluation
            logger.debug('expression "%s" of %s evaluated to %s', expression, port, json_utils.dumps(value))
            port.push_value(value, reason=core_ports.CHANGE_REASON_EXPRESSION)

        else:
            if changed_set_ids:  # Only log unchanged values if a port ID triggered evaluation
                logger.debug('%s value unchanged after expression evaluation', port)


def force_eval_expressions(port: core_ports.BasePort = None) -> None:
    logger.debug('forcing expression evaluation for %s', port or 'all ports')

    _force_eval_expression_ports.add(port)
    if port:
        port.reset_change_reason()


def enable_updating() -> None:
    global _updating_enabled

    if not _updating_enabled:
        logger.debug('enabling update mechanism')
        _updating_enabled = True


def disable_updating() -> None:
    global _updating_enabled

    if _updating_enabled:
        logger.debug('disabling update mechanism')
        _updating_enabled = False


def is_ready() -> bool:
    return _ready


def set_ready() -> None:
    global _ready

    logger.debug('ready')
    _ready = True


def uptime() -> float:
    return time.time() - _start_time


async def init() -> None:
    global _update_loop_task
    global loop

    loop = asyncio.get_event_loop()

    force_eval_expressions()
    _update_loop_task = loop.create_task(update_loop())


async def cleanup() -> None:
    if _update_loop_task:
        _update_loop_task.cancel()
        await _update_loop_task



#!/usr/bin/env python

from qtoggleserver import commands
from qtoggleserver.web import server as web_server


async def main() -> bool:
    await web_server.init()

    return True  # Run loop afterwards


def execute() -> None:
    commands.execute(main())


if __name__ == '__main__':
    execute()




PUT_SLAVE_DEVICES = {
    "type": "array",
    "items": {
        "type": "object"
    }
}

POST_SLAVE_DEVICES = {
    "type": "object",
    "properties": {
        "scheme": {
            "enum": ["http", "https"]
        },
        "host": {
            "type": "string",
            "maxLength": 256
        },
        "port": {
            "type": "integer",
            "min": 0,
            "max": 65535
        },
        "path": {
            "type": "string",
            "maxLength": 256
        },
        "admin_password": {
            "type": "string",
            "maxLength": 32
        },
        "admin_password_hash": {
            "type": "string",
            "minLength": 64,
            "maxLength": 64
        },
        "poll_interval": {
            "type": "number",
            "min": 0,
            "max": 86400
        },
        "listen_enabled": {
            "type": "boolean"
        }
    },
    "additionalProperties": False,
    "required": [
        "scheme",
        "host",
        "port",
        "path"
    ]
}

PATCH_SLAVE_DEVICE = {
    "type": "object",
    "properties": {
        "enabled": {
            "type": "boolean"
        },
        "poll_interval": {
            "type": "number",
            "min": 0,
            "max": 86400
        },
        "listen_enabled": {
            "type": "boolean"
        }
    },
    "additionalProperties": False
}

POST_SLAVE_DEVICE_EVENTS = {
    "type": "object",
    "properties": {
        "type": {
            "type": "string"
        },
        "params": {
            "type": "object"
        }
    },
    "additionalProperties": False,
    "required": [
        "type"
    ]
}

PATCH_DISCOVERED_DEVICE = {
    "type": "object",
    "properties": {
        "attrs": {
            "type": "object"
        }
    },
    "additionalProperties": False,
    "required": [
        "attrs"
    ]
}




import argparse
import asyncio
import logging.config
import os
import signal
import sys
import types

from typing import Any, Dict, Optional

from tornado import httpclient
from tornado import netutil

from qtoggleserver import persist
from qtoggleserver import slaves
from qtoggleserver import system
from qtoggleserver import version
from qtoggleserver import web
from qtoggleserver.core import device
from qtoggleserver.core import events
from qtoggleserver.core import history
from qtoggleserver.core import main
from qtoggleserver.core import ports
from qtoggleserver.core import reverse
from qtoggleserver.core import sessions
from qtoggleserver.conf import settings
from qtoggleserver.core import vports
from qtoggleserver.core import webhooks
from qtoggleserver.slaves import devices as slaves_devices
from qtoggleserver.utils import conf as conf_utils
from qtoggleserver.utils import logging as logging_utils
from qtoggleserver import peripherals  # This must be imported after core.ports


logger: Optional[logging.Logger] = None
options: Optional[types.SimpleNamespace] = None

_stopping = False


def parse_args() -> None:
    global options

    description = f'qToggleServer {version.VERSION}'
    epilog = None

    class VersionAction(argparse.Action):
        def __call__(self, *args, **kwargs) -> None:
            sys.stdout.write(f'{description}\n')
            sys.exit()

    parser = argparse.ArgumentParser(
        prog=sys.argv[0],
        description=description,
        epilog=epilog,
        add_help=False,
        formatter_class=argparse.RawTextHelpFormatter
    )

    parser.add_argument(
        '-c',
        help='specify the configuration file',
        type=str,
        dest='config_file'
    )

    parser.add_argument(
        '-h',
        help='print this help and exit',
        action='help',
        default=argparse.SUPPRESS
    )

    parser.add_argument(
        '-v',
        help='print program version and exit',
        action=VersionAction,
        default=argparse.SUPPRESS,
        nargs=0
    )

    options = parser.parse_args()


def init_settings() -> None:
    if options.config_file:
        try:
            parsed_config = conf_utils.config_from_file(options.config_file)

        except IOError as e:
            sys.stderr.write(f'failed to open config file "{options.config_file}": {e}\n')
            sys.exit(-1)

        except Exception as e:
            sys.stderr.write(f'failed to load config file "{options.config_file}": {e}\n')
            sys.exit(-1)

        settings.source = os.path.abspath(options.config_file)

    else:
        parsed_config = {}

    def_config = conf_utils.obj_to_dict(settings)
    def_config = conf_utils.config_from_dict(def_config)

    config = conf_utils.config_merge(def_config, parsed_config)
    config = conf_utils.config_to_dict(config)

    conf_utils.update_obj_from_dict(settings, config)


def init_logging() -> None:
    global logger

    settings.logging['disable_existing_loggers'] = False
    logging.config.dictConfig(settings.logging)

    # Add memory logs handler
    root_logger = logging.getLogger()
    main.memory_logs = logging_utils.FifoMemoryHandler(capacity=settings.logging['memory_logs_buffer_len'])
    root_logger.addHandler(main.memory_logs)

    logger = logging.getLogger('qtoggleserver')

    logger.info('hello!')
    logger.info('this is qToggleServer %s', version.VERSION)

    # We can't do this in init_settings() because we have no logging there
    if options.config_file:
        logger.info('using config from %s', options.config_file)

    else:
        logger.info('using default config')


def init_signals() -> None:
    loop = asyncio.get_event_loop()

    async def shutdown() -> None:
        global _stopping

        if _stopping:
            logger.error('interrupt signal received again')
            # TODO: commit suicide
            return

        _stopping = True

        logger.info('interrupt signal received')

        loop.stop()

    loop.add_signal_handler(signal.SIGINT, lambda: asyncio.create_task(shutdown()))
    loop.add_signal_handler(signal.SIGTERM, lambda: asyncio.create_task(shutdown()))


def handle_loop_exception(loop: asyncio.AbstractEventLoop, context: Dict[str, Any]) -> None:
    if isinstance(context.get('exception'), asyncio.CancelledError):
        return  # Ignore any cancelled errors

    loop.default_exception_handler(context)


async def init_loop() -> None:
    loop = asyncio.get_event_loop()

    loop.set_exception_handler(handle_loop_exception)


async def cleanup_loop() -> None:
    # Cancel all running tasks
    tasks = [t for t in asyncio.all_tasks() if t is not asyncio.current_task()]

    for task in tasks:
        logger.debug(f'cancelling {task.get_coro()}')
        task.cancel()

    await asyncio.gather(*tasks)

    await asyncio.get_event_loop().shutdown_asyncgens()


def init_tornado() -> None:
    httpclient.AsyncHTTPClient.configure('tornado.simple_httpclient.SimpleAsyncHTTPClient', max_clients=1024)
    netutil.Resolver.configure(
        'tornado.netutil.OverrideResolver',
        mapping=system.dns.get_custom_dns_mapping_dict(),
        resolver=netutil.DefaultExecutorResolver()
    )


async def init_persist() -> None:
    logger.info('initializing persistence')
    try:
        await persist.get_value('device')

    except Exception as e:
        logger.error('failed to initialize persistence: %s', e, exc_info=True)
        sys.exit(-1)


async def cleanup_persist() -> None:
    logger.info('cleaning up persistence')
    await persist.cleanup()


async def init_system() -> None:
    logger.info('initializing system')
    await system.init()


async def cleanup_system() -> None:
    logger.info('cleaning up system')
    await system.cleanup()


async def init_events() -> None:
    logger.info('initializing events')
    await events.init()


async def cleanup_events() -> None:
    logger.info('cleaning up events')
    await events.cleanup()


async def init_sessions() -> None:
    logger.info('initializing sessions')
    await sessions.init()


async def cleanup_sessions() -> None:
    logger.info('cleaning up sessions')
    await sessions.cleanup()


async def init_history() -> None:
    if history.is_enabled():
        logger.info('initializing history')
        await history.init()


async def cleanup_history() -> None:
    if history.is_enabled():
        logger.info('cleaning history')
        await history.cleanup()


async def init_device() -> None:
    logger.info('initializing device')
    await device.init()


async def cleanup_device() -> None:
    logger.info('cleaning up device')
    await device.cleanup()


async def init_webhooks() -> None:
    if settings.webhooks.enabled:
        logger.info('initializing webhooks')
        await webhooks.init()


async def cleanup_webhooks() -> None:
    if settings.webhooks.enabled:
        logger.info('cleaning up webhooks')
        await webhooks.cleanup()


async def init_reverse() -> None:
    if settings.reverse.enabled:
        logger.info('initializing reverse API calls')
        await reverse.init()


async def cleanup_reverse() -> None:
    if settings.reverse.enabled:
        logger.info('cleaning up reverse API calls')
        await reverse.cleanup()


async def init_ports() -> None:
    logger.info('initializing ports')

    # Load ports statically configured in settings
    await ports.init()

    # Peripheral ports
    for peripheral in peripherals.all_peripherals():
        try:
            port_args = await peripheral.get_port_args()
            # Use raise_on_error=False because we prefer a partial successful startup rather than a failed one
            loaded_ports = await ports.load(port_args, raise_on_error=False)
            peripheral.set_ports(loaded_ports)

        except Exception as e:
            logger.error('failed to load ports of %s: %s', peripheral, e, exc_info=True)

    # Load virtual ports
    await vports.init()


async def cleanup_ports() -> None:
    logger.info('cleaning up ports')
    await ports.cleanup()


async def init_slaves() -> None:
    if settings.slaves.enabled:
        logger.info('initializing slaves')
        await slaves.init()


async def cleanup_slaves() -> None:
    if settings.slaves.enabled:
        logger.info('cleaning up slaves')
        await slaves.cleanup()


async def init_peripherals() -> None:
    logger.info('initializing peripherals')
    await peripherals.init()


async def cleanup_peripherals() -> None:
    logger.info('cleaning up peripherals')
    await peripherals.cleanup()


async def init_main() -> None:
    logger.info('initializing main')
    await main.init()

    # Wait until slaves are also ready before actually considering main loop ready
    if settings.slaves.enabled:
        logger.debug('waiting for slaves to become ready')
        while not slaves_devices.ready():
            await asyncio.sleep(0.1)

        logger.debug('slaves are ready')

    # Mark main as ready after all slaves with their ports have been initialized and hopefully brought online. Allow an
    # extra second for pending loop tasks.
    await asyncio.sleep(1)
    main.set_ready()


async def cleanup_main() -> None:
    logger.info('cleaning up main')
    await main.cleanup()


async def init_web() -> None:
    logger.info('initializing web')
    await web.init()


async def cleanup_web() -> None:
    logger.info('cleaning up web')
    await web.cleanup()


async def init() -> None:
    parse_args()

    init_settings()
    init_logging()
    init_signals()
    init_tornado()

    await init_system()
    await init_persist()
    await init_peripherals()
    await init_events()
    await init_sessions()
    await init_history()
    await init_device()
    await init_webhooks()
    await init_reverse()
    await init_ports()
    await init_slaves()
    await init_main()
    await init_web()


async def cleanup() -> None:
    await cleanup_web()
    await cleanup_main()
    await cleanup_ports()
    await cleanup_slaves()
    await cleanup_reverse()
    await cleanup_webhooks()
    await cleanup_device()
    await cleanup_history()
    await cleanup_sessions()
    await cleanup_events()
    await cleanup_peripherals()
    await cleanup_persist()
    await cleanup_system()




from __future__ import annotations

import inspect
import asyncio
import re

from typing import Any, Callable, Optional

from qtoggleserver.conf import settings
from qtoggleserver.core import api as core_api
from qtoggleserver.core import events as core_events
from qtoggleserver.core import main as core_main
from qtoggleserver.core import responses as core_responses
from qtoggleserver.core.api import auth as core_api_auth
from qtoggleserver.core.api import schema as core_api_schema
from qtoggleserver.core.typing import GenericJSONDict, GenericJSONList
from qtoggleserver.slaves import devices as slaves_devices
from qtoggleserver.slaves import exceptions as slaves_exceptions

from .. import schema as api_schema


_LONG_TIMEOUT_API_CALLS = [
    ('PATCH', re.compile(r'/device/?')),
    ('GET', re.compile(r'/firmware/?')),
    ('PATCH', re.compile(r'/firmware/?')),
    ('PATCH', re.compile(r'/ports/[^/]+/?')),
    ('PATCH', re.compile(r'/ports/[^/]+/value/?')),
    ('POST', re.compile(r'/devices/?')),
    ('PATCH', re.compile(r'/devices/[^/]+/?'))
]


async def add_slave_device(properties: GenericJSONDict) -> slaves_devices.Slave:
    properties = dict(properties)  # Work on copy, don't mess up incoming argument

    scheme = properties.pop('scheme')
    host = properties.pop('host')
    port = properties.pop('port')
    path = properties.pop('path')
    admin_password = properties.pop('admin_password', None)
    admin_password_hash = properties.pop('admin_password_hash', None)
    poll_interval = properties.pop('poll_interval', 0)
    listen_enabled = properties.pop('listen_enabled', None)

    # Look for slave duplicate
    for slave in slaves_devices.get_all():
        if (slave.get_scheme() == scheme and
            slave.get_host() == host and
            slave.get_port() == port and
            slave.get_path() == path):

            raise core_api.APIError(400, 'duplicate-device')

    if poll_interval and listen_enabled:
        raise core_api.APIError(400, 'listening-and-polling')

    # Ensure admin password is supplied, in a way or another
    if admin_password is None and admin_password_hash is None:
        raise core_api.APIError(400, 'missing-field', field='admin_password')

    try:
        slave = await slaves_devices.add(
            scheme,
            host,
            port,
            path,
            poll_interval,
            listen_enabled,
            admin_password=admin_password,
            admin_password_hash=admin_password_hash,
            **properties
        )

    except (core_responses.HostUnreachable,
            core_responses.NetworkUnreachable,
            core_responses.UnresolvableHostname) as e:

        raise core_api.APIError(502, 'unreachable') from e

    except core_responses.ConnectionRefused as e:
        raise core_api.APIError(502, 'connection-refused') from e

    except core_responses.InvalidJson as e:
        raise core_api.APIError(502, 'invalid-device') from e

    except core_responses.Timeout as e:
        raise core_api.APIError(504, 'device-timeout') from e

    except slaves_exceptions.InvalidDevice as e:
        raise core_api.APIError(502, 'invalid-device') from e

    except slaves_exceptions.NoListenSupport as e:
        raise core_api.APIError(400, 'no-listen-support') from e

    except slaves_exceptions.DeviceAlreadyExists as e:
        raise core_api.APIError(400, 'duplicate-device') from e

    except core_api.APIError:
        raise

    except core_responses.HTTPError as e:
        # We need to treat the 401/403 slave responses as a 400
        if e.code in (401, 403):
            raise core_api.APIError(400, 'forbidden') from e

        raise core_api.APIError.from_http_error(e) from e

    except Exception as e:
        raise slaves_exceptions.adapt_api_error(e) from e

    return slave


async def add_slave_device_retry_disabled(properties: GenericJSONDict) -> slaves_devices.Slave:
    try:
        return await add_slave_device(properties)

    except core_api.APIError:
        if properties.get('enabled', True):
            core_api.logger.warning('adding device failed, adding it as disabled', exc_info=True)
            return await add_slave_device(dict(properties, enabled=False))

        else:
            raise


async def wrap_error_with_index(index: int, func: Callable, *args, **kwargs) -> Any:
    try:
        result = func(*args, **kwargs)
        if inspect.isawaitable(result):
            result = await result

    except core_api.APIError as e:
        raise core_api.APIError(
            status=e.status,
            code=e.code,
            index=index,
            **e.params
        )

    except asyncio.TimeoutError:
        raise core_api.APIError(
            status=504,
            code='device-timeout',
            index=index
        )

    except Exception as e:
        raise core_api.APIError(
            status=500,
            code='unexpected-error',
            message=str(e),
            index=index
        )

    return result


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
async def get_slave_devices(request: core_api.APIRequest) -> GenericJSONList:
    return [slave.to_json() for slave in slaves_devices.get_all()]


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
async def put_slave_devices(request: core_api.APIRequest, params: GenericJSONList) -> None:
    if not settings.core.backup_support:
        raise core_api.APIError(404, 'no-such-function')

    core_api_schema.validate(
        params,
        api_schema.PUT_SLAVE_DEVICES
    )

    core_api.logger.debug('restoring slave devices')

    # Disable event handling during the processing of this request, as we're going to trigger a full-update at the end
    core_events.disable()

    # Temporarily disable core updating (port polling, expression evaluating and value-change handling)
    core_main.disable_updating()

    try:
        # Remove all slave devices
        for slave in slaves_devices.get_all():
            await slaves_devices.remove(slave)

        add_device_schema = dict(api_schema.POST_SLAVE_DEVICES)
        add_device_schema['additionalProperties'] = True

        # Validate supplied slave properties
        for index, properties in enumerate(params):
            await wrap_error_with_index(
                index,
                core_api_schema.validate,
                properties,
                add_device_schema
            )

        # Add slave devices
        add_slave_futures = []
        for index, properties in enumerate(params):
            add_slave_future = wrap_error_with_index(
                index,
                add_slave_device_retry_disabled,
                properties
            )
            add_slave_futures.append(add_slave_future)

        added_slaves = await asyncio.gather(*add_slave_futures)

        # Wait for devices to come online
        wait_online_futures = []
        for slave in added_slaves:
            if slave.is_enabled():
                wait_online_futures.append(slave.wait_online(timeout=settings.slaves.long_timeout))

        try:
            await asyncio.gather(*wait_online_futures)

        except asyncio.TimeoutError:
            # Ignore timeouts; this is a best-effort API call
            pass

    finally:
        core_main.enable_updating()
        core_events.enable()

    await core_events.trigger_full_update()

    core_api.logger.debug('slave devices restore done')


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
async def post_slave_devices(request: core_api.APIRequest, params: GenericJSONDict) -> GenericJSONDict:
    core_api_schema.validate(params, api_schema.POST_SLAVE_DEVICES)
    slave = await add_slave_device(params)

    return slave.to_json()


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
async def patch_slave_device(request: core_api.APIRequest, name: str, params: GenericJSONDict) -> None:
    core_api_schema.validate(params, api_schema.PATCH_SLAVE_DEVICE)

    slave = slaves_devices.get(name)
    if not slave:
        raise core_api.APIError(404, 'no-such-device')

    if params.get('enabled') is True and not slave.is_enabled():
        await slave.enable()

    elif params.get('enabled') is False and slave.is_enabled():
        await slave.disable()

    if params.get('poll_interval') and params.get('listen_enabled'):
        raise core_api.APIError(400, 'listening-and-polling')

    if params.get('poll_interval') is not None:
        slave.set_poll_interval(params['poll_interval'])

    if params.get('listen_enabled') is not None:
        if params['listen_enabled']:
            # We need to know if device supports listening; we therefore call GET /device before enabling it

            if slave.is_enabled():
                try:
                    attrs = await slave.api_call('GET', '/device')

                except Exception as e:
                    raise slaves_exceptions.adapt_api_error(e) from e

                if 'listen' not in attrs['flags']:
                    raise core_api.APIError(400, 'no-listen-support')

            slave.enable_listen()

        else:
            slave.disable_listen()

    await slave.save()
    await slave.trigger_update()


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
async def slave_device_forward(
    request: core_api.APIRequest,
    name: str,
    path: str,
    params: Optional[GenericJSONDict] = None,
    internal_use: bool = False
) -> Any:

    slave = slaves_devices.get(name)

    if not slave:
        raise core_api.APIError(404, 'no-such-device')

    if not internal_use:
        if not path.startswith('/'):
            path = '/' + path

        if path.startswith('/listen'):
            raise core_api.APIError(404, 'no-such-function')

    intercepted, response = await slave.intercept_request(request.method, path, params, request)
    if intercepted:
        return response

    override_disabled = request.query.get('override_disabled')
    if not slave.is_enabled() and (override_disabled != 'true'):
        raise core_api.APIError(404, 'device-disabled')

    override_offline = request.query.get('override_offline')
    if (not slave.is_online() or not slave.is_ready()) and (override_offline != 'true'):
        raise core_api.APIError(503, 'device-offline')

    timeout = request.query.get('timeout')
    if timeout is not None:
        try:
            timeout = int(timeout)

        except ValueError:
            raise core_api.APIError(400, 'invalid-field', field='timeout') from None

    else:
        # Use default slave timeout unless API call requires longer timeout
        timeout = settings.slaves.timeout
        for m, path_re in _LONG_TIMEOUT_API_CALLS:
            if request.method == m and path_re.fullmatch(path):
                timeout = settings.slaves.long_timeout
                break

    try:
        response = await slave.api_call(request.method, path, params, timeout=timeout, retry_counter=None)

    except Exception as e:
        raise slaves_exceptions.adapt_api_error(e) from e

    return response


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
async def delete_slave_device(request: core_api.APIRequest, name: str) -> None:
    slave = slaves_devices.get(name)
    if not slave:
        raise core_api.APIError(404, 'no-such-device')

    await slaves_devices.remove(slave)


@core_api.api_call(core_api.ACCESS_LEVEL_NONE)
async def post_slave_device_events(request: core_api.APIRequest, name: str, params: GenericJSONDict) -> None:
    slave = slaves_devices.get(name)
    if not slave:
        raise core_api.APIError(404, 'no-such-device')

    # Slave events endpoint has special privilege requirements: its token signature must be validated using slave admin
    # password

    auth = request.headers.get('Authorization')
    if not auth:
        slave.warning('missing authorization header')
        raise core_api.APIError(401, 'authentication-required')

    try:
        core_api_auth.parse_auth_header(
            auth,
            core_api_auth.ORIGIN_DEVICE,
            lambda u: slave.get_admin_password_hash(),
            require_usr=False
        )

    except core_api_auth.AuthError as e:
        slave.warning(str(e))
        raise core_api.APIError(401, 'authentication-required') from e

    core_api_schema.validate(params, api_schema.POST_SLAVE_DEVICE_EVENTS)

    if slave.get_poll_interval() > 0:
        raise core_api.APIError(400, 'polling-enabled')

    if slave.is_listen_enabled():
        raise core_api.APIError(400, 'listening-enabled')

    # At this point we can be sure the slave is permanently offline

    try:
        await slave.handle_event(params)

    except Exception as e:
        raise core_api.APIError(500, 'unexpected-error', message=str(e)) from e

    slave.update_last_sync()
    await slave.save()

    # As soon as we receive event requests (normally generated by a webhook), we can consider the device is temporarily
    # reachable, so we apply provisioning values and update data locally
    slave.schedule_provisioning_and_update(1)




from __future__ import annotations

import asyncio
import ctypes
import logging
import psutil
import random
import re
import socket
import struct
import time

from typing import List, Optional


DEFAULT_TIMEOUT = 10
MAX_TIMEOUT = 60

ETH_P_ALL = 3
IFF_PROMISC = 0x100
SIOCGIFFLAGS = 0x8913
SIOCSIFFLAGS = 0x8914

DHCP_MESSAGETYPE = 0x35
DHCP_EOF = 0xFF
DHCP_MSGDISCOVER = 0x01
DHCP_HOSTNAME = 0x0c
DHCP_REQUEST = 0x1
DHCP_REPLY = 0x2
DHCP_MAGIC = 0x63825363
DHCP_SRC_PORT = 68
DHCP_DST_PORT = 67

IPV4_PROTO_UDP = 17

ETHERTYPE_IP = 0x0800
ARPHRD_ETHER = 1

logger = logging.getLogger(__name__)

_dhcp_replies: List[DHCPReply] = []


class DHCPException(Exception):
    pass


class DHCPTimeout(DHCPException):
    pass


class DHCPReply:
    def __init__(
        self,
        ip_address: str,
        xid: int,
        timestamp: float
    ) -> None:

        self.ip_address: str = ip_address
        self.xid: int = xid
        self.timestamp: float = timestamp

    def __str__(self) -> str:
        return f'DHCP offer for {self.ip_address}'


class _ifreq(ctypes.Structure):
    _fields_ = [
        ("ifr_ifrn", ctypes.c_char * 16),
        ("ifr_flags", ctypes.c_short)
    ]


def _build_dhcp_options(hostname: Optional[str]) -> bytes:
    dhcp_opts = b''
    dhcp_opts += struct.pack('BBB', DHCP_MESSAGETYPE, 1, DHCP_MSGDISCOVER)  # Message type
    if hostname:
        dhcp_opts += struct.pack('BB', DHCP_HOSTNAME, len(hostname))
        dhcp_opts += hostname.encode()
    dhcp_opts += struct.pack('B', DHCP_EOF)  # EOF

    return dhcp_opts


def _build_dhcp_header(mac_address: str, xid: int, own_ip_address: str) -> bytes:
    own_ip_address = [int(p) for p in own_ip_address.split('.')]

    dhcp_header = struct.pack('B', DHCP_REQUEST)  # Opcode
    dhcp_header += struct.pack('B', ARPHRD_ETHER)  # Hardware address type
    dhcp_header += struct.pack('B', 6)  # Hardware address len
    dhcp_header += struct.pack('B', 0)  # Opcount
    dhcp_header += struct.pack('!L', xid)  # Transaction ID
    dhcp_header += struct.pack('!H', 0)  # Seconds
    dhcp_header += struct.pack('!H', 0)  # Flags
    dhcp_header += struct.pack('BBBB', 0, 0, 0, 0)  # Client IP
    dhcp_header += struct.pack('BBBB', 0, 0, 0, 0)  # Your IP
    dhcp_header += struct.pack('BBBB', 0, 0, 0, 0)  # Server IP
    dhcp_header += struct.pack('BBBB', *own_ip_address)  # Gateway IP
    dhcp_header += bytes.fromhex(re.sub('[^a-fA-F0-9]', '', mac_address))  # Client MAC
    dhcp_header += b'\x00' * 10  # Remaining hardware address space
    dhcp_header += b'\x00' * 64  # Server hostname
    dhcp_header += b'\x00' * 128  # Boot file name
    dhcp_header += struct.pack('!L', DHCP_MAGIC)

    return dhcp_header


def _build_udp_header(dhcp_packet: bytes) -> bytes:
    udp_header = struct.pack('!H', DHCP_SRC_PORT)  # Source port
    udp_header += struct.pack('!H', DHCP_DST_PORT)  # Dest port
    udp_header += struct.pack('!H', 8 + len(dhcp_packet))  # UDP header + DHCP packet
    udp_header += struct.pack('!H', 0)  # Checksum

    return udp_header


def _build_ipv4_header(udp_packet: bytes) -> bytes:
    ipv4_header = b'\x45'   # Version + Length
    ipv4_header += b'\x00'  # ToS
    ipv4_header += struct.pack('!H', 20 + len(udp_packet))  # IP header + UDP packet
    ipv4_header += struct.pack('!H', 0)  # ID
    ipv4_header += struct.pack('!H', 0)  # Fragment offset
    ipv4_header += struct.pack('B', 64)  # TTL
    ipv4_header += struct.pack('B', IPV4_PROTO_UDP)  # Protocol
    ipv4_header += struct.pack('!H', 0)  # Checksum
    ipv4_header += struct.pack('BBBB', 0, 0, 0, 0)  # Source address
    ipv4_header += struct.pack('BBBB', 255, 255, 255, 255)  # Dest address

    ipv4_checksum = 0
    for i in range(6):  # Header length is 10x16 bytes
        word = struct.unpack('!H', ipv4_header[i * 2: i * 2 + 2])[0]
        ipv4_checksum += word

    ipv4_checksum = (ipv4_checksum >> 16) + ipv4_checksum
    ipv4_checksum = (~ipv4_checksum) & 0xFFFF

    ipv4_header = ipv4_header[0:10] + struct.pack('!H', ipv4_checksum) + ipv4_header[12:20]

    return ipv4_header


def _build_ethernet_header(own_mac_address: str) -> bytes:
    ethernet_header = b'\xFF\xFF\xFF\xFF\xFF\xFF'  # Dest MAC
    ethernet_header += bytes.fromhex(re.sub('[^a-fA-F0-9]', '', own_mac_address))
    ethernet_header += struct.pack('!H', ETHERTYPE_IP)

    return ethernet_header


def _check_received_frame(frame: bytes, expected_xid: int) -> Optional[DHCPReply]:
    if len(frame) < 282:
        return

    ethernet_header = frame[:14]
    ipv4_header = frame[14:34]
    udp_header = frame[34:42]
    dhcp_header = frame[42:282]
    # dhcp_opts = frame[282:]

    ethertype = struct.unpack('!H', ethernet_header[12:14])[0]
    if ethertype != ETHERTYPE_IP:
        return

    if ipv4_header[9] != IPV4_PROTO_UDP:
        return

    src_port = struct.unpack('!H', udp_header[0:2])[0]
    if src_port != DHCP_DST_PORT:
        return

    dst_port = struct.unpack('!H', udp_header[2:4])[0]
    if (dst_port != DHCP_SRC_PORT) and (dst_port != DHCP_DST_PORT):
        return

    # At this point we can consider we're dealing with a DHCP reply

    xid = struct.unpack('!L', dhcp_header[4:8])[0]
    your_ip = dhcp_header[16:20]
    your_ip = '{:d}.{:d}.{:d}.{:d}'.format(*your_ip)
    dhcp_reply = DHCPReply(
        ip_address=your_ip,
        xid=xid,
        timestamp=time.time()
    )

    if expected_xid == xid:
        return dhcp_reply

    else:  # Not ours, queue it for other pending requests
        _dhcp_replies.append(dhcp_reply)


def _check_and_prune_pending_dhcp_replies(expected_xid: int) -> Optional[DHCPReply]:
    global _dhcp_replies

    # Prune expired replies
    now = time.time()
    _dhcp_replies = [r for r in _dhcp_replies if now - r.timestamp < MAX_TIMEOUT]

    # Look for our transaction id
    reply = next((r for r in _dhcp_replies if r.xid == expected_xid), None)
    if reply:
        _dhcp_replies.remove(reply)

    return reply


async def request(
    interface: str,
    mac_address: str,
    hostname: Optional[str] = None,
    timeout: int = DEFAULT_TIMEOUT
) -> DHCPReply:

    logger.debug(
        'sending DHCP discovery for %s (%s) on %s',
        mac_address,
        hostname if hostname else '<no hostname>',
        interface
    )

    # Generate random transaction id
    xid = int(random.random() * 0xFFFFFFFF)

    # Find MAC & IP address of given interface
    try:
        if_addrs = psutil.net_if_addrs()[interface]

    except KeyError:
        raise DHCPException(f'Cannot find own address for interface {interface}')

    own_mac_address = None
    own_ip_address = None
    for addr in if_addrs:
        if addr.family == socket.AF_PACKET:
            own_mac_address = addr.address

        elif addr.family == socket.AF_INET:
            own_ip_address = addr.address

    if not own_ip_address or not own_mac_address:
        raise DHCPException(f'Cannot find own address for interface {interface}')

    dhcp_opts = _build_dhcp_options(hostname)
    dhcp_header = _build_dhcp_header(mac_address, xid, own_ip_address)
    dhcp_packet = dhcp_header + dhcp_opts

    udp_header = _build_udp_header(dhcp_packet)
    udp_packet = udp_header + dhcp_packet

    ipv4_header = _build_ipv4_header(udp_packet)
    ipv4_packet = ipv4_header + udp_packet

    ethernet_header = _build_ethernet_header(own_mac_address)
    ethernet_frame = ethernet_header + ipv4_packet

    # Open sockets
    sock = socket.socket(family=socket.PF_PACKET, type=socket.SOCK_RAW, proto=socket.htons(ETH_P_ALL))
    sock.bind((interface, 0))

    # Actually send packet
    sock.send(ethernet_frame)

    # Wait for data; filter out stuff that we don't need
    start_time = time.time()
    while True:
        try:
            data = sock.recv(1024, socket.MSG_DONTWAIT)

        except BlockingIOError:
            data = None

        if data:
            offer = _check_received_frame(data, xid)
            if offer:
                break

        offer = _check_and_prune_pending_dhcp_replies(xid)
        if offer:
            break

        if time.time() - start_time > timeout:
            break

        await asyncio.sleep(0)

    # Don't forget to close socket
    sock.close()

    if offer:
        logger.debug(
            'received DHCP offer for %s (%s) on %s: %s',
            mac_address,
            hostname if hostname else '<no hostname>',
            interface,
            offer.ip_address
        )

        return offer

    else:
        logger.warning(
            'timeout waiting for DHCP offer for %s (%s) on %s',
            mac_address,
            hostname if hostname else '<no hostname>',
            interface
        )

        raise DHCPTimeout()




from .base import Evaluated
from .functions import function, Function


@function('ABS')
class AbsFunction(Function):
    MIN_ARGS = MAX_ARGS = 1

    async def eval(self) -> Evaluated:
        return abs((await self.eval_args())[0])


@function('SGN')
class SgnFunction(Function):
    MIN_ARGS = MAX_ARGS = 1

    async def eval(self) -> Evaluated:
        e = int((await self.eval_args())[0])
        if e > 0:
            return 1

        elif e < 0:
            return -1

        else:
            return 0
